/*
  RabidVWM LiteStep Module - version 1.0
	Main Header File

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
 
#include "stdafx.h"

// AggressiveOptimize from www.nopcode.com
#include "..\AggressiveOptimize.h"
#include "old\lsapi.h"

#ifndef NDEBUG
#define TRACE_FILE "C:\\RABIDVWMLOG.HTML"
#define VWM_STATS
#endif

// not enabled because they're pretty useless
//#define INCLUDE_SCROLL_EVENTS
// not completely implemented yet
//#define INCLUDE_DROP_EVENTS
#define RABIDVWM_MULTIMON
#define VWMDROPTARGET
#define USE_PARSECOORDINATE

#ifndef LSTOOLS_REV
#define NO_RCMM
#endif

const char szAppName[] = "RabidVWM";
#ifndef NDEBUG
#define BUILD_TYPE " d"
#else
#define BUILD_TYPE " r"
#endif
#if !defined(BUILD_ID) && defined(NDEBUG)
#define BUILD_ID
#endif
#ifndef BUILD_ID
#define BUILD_ID ", built " __DATE__ " at " __TIME__
#endif

const char szAppVer[] = "1.0" BUILD_TYPE BUILD_ID;
const char szAppAuth[] = "RabidCow (rabidcow@juno.com)";

struct StarWindowMatchItem {
	StarWindowMatchItem *pNext;

	char *Key;
	DWORD MatchProcFlag; // each bit corresponds to an item in the g_pWhdStringProc array
	char *Args;
};

struct s_WinData;
typedef void StarDataProc(s_WinData *pData, StarWindowMatchItem *pMatch);
typedef HWND StarGetWndProc(s_WinData *pData);
typedef void WinModifier(s_WinData *pData,DWORD lParam);
typedef COLORREF GetPixelProc(HDC src, COLORREF back, int x, int y);

struct StarWindowMatchList {
	const char *Key;

	StarWindowMatchItem *pItem;

	StarDataProc *Set;
	StarDataProc *Clear;
	StarGetWndProc *GetH;
};

typedef void MouseProc(int x,int y);

// an easy and fast way to select mouse actions independent of button
struct s_MouseAction {
	MouseProc *Press;
	MouseProc *Move;
	MouseProc *Release;
	MouseProc *Cancel;
};

struct s_ParseData {
	// more than I actually use at this point, but hey,
	// it's only taking memory during ReadConfig.

	char line[MAX_LINE_LENGTH];
	char one[MAX_PATH_LENGTH],
			 two[MAX_PATH_LENGTH],
			 three[MAX_PATH_LENGTH],
			 four[MAX_PATH_LENGTH],
			 five[MAX_PATH_LENGTH];
	char *tokens[5];
	char keybuf[256];
};

class PixelBlock;

struct s_PaintWnd {

//	void (*PaintVWM)(HDC dc); --> chosen separately from the others
	// called to paint the main VWM window
	// I might want a mode that draws by desk,
	// with another to draw the select...

	void (*BeginPaintWnd)(HDC dc,PixelBlock &,void *store);
	// called to initialize the miniwindow drawing loop
	// store is a chunk of memory, PAINTWND_BUFFER_SIZE bytes,
	// passed to all functions in the loop.  Use it wisely,
	// and make sure it's big enough.

	void (*PaintWnd)(HDC dc,PixelBlock &,void *store,s_WinData *pData,RECT *rFull,RECT *rWnd);
	// paint the main window
	// if titles are off, rFull will == rWnd
	// otherwise rWnd does not include the title area
	// DO NOT draw icons in this routine, that's separate

	void (*PaintWndTitle)(HDC dc,PixelBlock &,void *store,s_WinData *pData,RECT *rFull,RECT *rTitle);
	// paint the window title
	// this is only called if bTitlebars is true
	// rFull = full miniwindow, rTitle = title area

	// the shared icon drawing code goes here

	void (*EndPaintWnd)(HDC dc,PixelBlock &,void *store);
	// called to clean up the miniwindow drawing loop
};

struct s_WinData {
	HWND hWnd;        // used as a key to find this data
	HICON hIcon;
	int LastIconSize; // to verify that the cached hIcon will still work
	RECT Area;        // screen area
	RECT OldArea;     // old screen area to detect move/resize
	RECT Mini;        // vwm area
	DWORD TitleChecksum; // used to detect title changes

	HWND Prev;        // next window down in paint order
	int PaintNext;    // next window up in paint order

	int Alpha;        // I don't quite remember why I wanted to let windows have independent alpha

	int HomeDesk;     // desk which this window calls home

	bool bTouch;      // detect removed windows
	bool bHidden;     // windows that are hidden but should still be drawn
	bool bTitleChanged;
	bool bSizeChanged;
	bool bPosChanged;
	bool bOrderChanged;
	bool bFlash;      // true if window is flashing (ie, this flag does not "flash")
	bool bFlashOn;		// true if the window should stick in the flash state
	unsigned iFlashCountLeft; // number of times left before sticking on

	bool bNewWindow;
	bool bWasDrawn;   // track ShouldDrawWnd on removal

	bool bIgnoreTitleChange;
	bool bSticky;
	bool bFix;

	HBITMAP hSnapShot;
	HMONITOR hMonitor;
};

struct s_DeskData {
	HWND hLastActiveWnd;
	int nWnds;        // the sole purpose of this var is for ^@de^
	                  // which seems to me like it should be pretty useful
};

struct s_BevelColors {
	COLORREF Flat,Light,Dark;
	COLORREF Border;
	int Alpha;
};

#ifdef LSTOOLS_REV
typedef lsEdgeMap s_EdgeMap;
#else
struct s_EdgeMap {
	HBITMAP Bitmap;
	SIZE Size;
	RECT Edge;
	bool Tile;
};
#endif

struct s_Action {
	const char *Command;
	s_WinData *pData;
	POINT MousePos;
	int Hold;
#ifdef VWM_STATS
	DWORD nCalled;
	DWORD nActivated;
	DWORD tActivated;
#endif
};

#ifdef VWM_STATS
struct s_VWMStats {
	DWORD InitStart;
	DWORD InitTime;
	DWORD WndProcStart;
	DWORD WndProcTime;
	DWORD WndProcLevel;
	DWORD CreateTime;
	DWORD RefreshTime;
	DWORD TraceTime;
	DWORD EdgeBltTime;
	DWORD UpdateTime;
	DWORD SnapShotTime;
	DWORD PaintTime;
	DWORD StarListTime;

	long GDIBrushs;
	long GDIPens;
	long GDIDCs;
	long GDIBitmaps;
};

extern volatile s_VWMStats VWMStats;

#define StatGDIAdd(type) { ++VWMStats.GDI##type##s; }
#define StatGDIDel(type) { ASSERT(VWMStats.GDI##type##s>0); --VWMStats.GDI##type##s; }
#else /* VWM_STATS */ 

#define StatGDIAdd(type) (void)0
#define StatGDIDel(type) (void)0
#endif /* VWM_STATS */ 

#ifdef VWMDROPTARGET
class VWMDropTarget;   // our IDropTarget interface
#endif

struct s_GlobalData {

	s_WinData *KnownWindows;
	int nKnownWindows;   // number of windows that the vwm needs to deal with
	int nMaxWindows;     // capacity of KnownWindows list
	int nEnumWindows;    // number of windows enumerated in the last pass
	                     // this should be == nKnownWindows unless a window has been
	                     // removed, which is taken care of in UpdateWindowList
	int FirstWindow;     // in paint order (bottom-most window)

	s_WinData *pRemember;// window being dragged, needs to remain valid across updates

	s_DeskData *Desks;
	int nDesks;          // total number of desks
	int nDesksX;         // number of desks across
	int nDesksY;         // number of desks down
	int CurrentDesk;
	SIZE VWMDeskSize;    // size of a desktop in VWM pixels

	HMONITOR hMonitorToSwitch; // need to change this to a list

	RECT rcDesk;         // area windows will be in
	RECT rcScreen;       // area VWM window will be in

	SIZE DeskSize;       // screen size + margin
	POINT DeskOffset;    // offset to make all desks align to 0,0 plus a multiple of the desktop size

	SIZE ScreenSize;     // total size of switched monitors
	POINT ScreenOffset;  // top,left corner of screen (not always 0,0)
	SIZE Scale;          // real desktop size / vwm desktop size

	const s_PaintWnd *Painter; // set of functions for the miniwindow paint loop
	void (*PaintVWM)(HDC dc);  // called to paint the main VWM window
	void (*PaintSelect)(HDC,s_DeskData *,RECT *); // paint the selected desktop	

	bool bWindowTouch;   // to recognize when a window is not enumerated, ie goes away
	bool bFlashState;    // coordinates flashing so flashes are all in sync, this toggles on each repaint
	bool bChangeState;   // set if we need to redraw
	bool bChangeOrder;   // some windows have changed z-order
	bool bDelaySnapshot; // tells snapshot mode it needs to try again in a bit

	HWND hParentWnd;     // parent of main window when not always-on-top
	HWND hMainWnd;       // main window itself, set in WM_CREATE
	HWND hLiteStepWnd;   // cached from GetLitestepWnd()

	HWND hFocusWnd;      // window with input focus
	HWND hRootFocusWnd;  // root owner of window with input focus
	HWND hLastEnumWnd;   // outside of UpdateWindowListProc, topmost window
	                     // otherwise used to capture paint order
	HWND hWndDragging;   // don't move this window while switching desks

	HINSTANCE hInstance;

// timer values, random unique numbers
#define TIMER_REFRESH 10734
	unsigned int RefreshTimerID;  // main timer for window updates
#define TIMER_HOVER 11305
	unsigned int HoverTimerID;    // keep track of mouse hovering
#define TIMER_FOCUS 9983
	unsigned int FocusTimerID;    // used to force focus if the first try didn't work
#define TIMER_SNAPSHOT 8760
	unsigned int SnapShotTimerID; // used for delayed snapshot rendering

	int RefreshTime;

	bool bMouseHovering;
	bool bMouseOver;
	POINT HoverPoint;
	int HoverTime;

#ifdef VWMDROPTARGET
	bool bInDragDrop;    // to trigger switch-hover instead of normal hover behavior
#endif

	unsigned int MouseAction[3]; // index into MouseTable[4]
	                             // one for each button: left, right, middle
	                             // add more for more buttons if desired
	int MouseCurrent; // which button-action is being processed, ** 1-based **
	void *MouseActionData;

	s_BevelColors cBack;   // vwm window
	s_BevelColors cSelect; // selected desk color
	s_BevelColors cWnd;    // miniwindow background
	s_BevelColors cTitle;  // miniwindow title
	s_BevelColors cFlash;  // flash miniwindow background
	int DoubleBevelWidth;  // size of the border on windows for bevel

	unsigned iFlashCount;  // number of times to flash a window

#ifdef VWMREALTRANS
	bool bWin2kTrans;      // alternate transparency model
#endif

	POINT VWMPos;
	SIZE VWMSize;
	HBITMAP bmBuffer;			// complete VWM image
	HBITMAP bmTransparent;// desktop behind the VWM
// not used, were probably for caching layers when only one
// needed to change, but I never got around to doing that
//	HBITMAP bmLayer_Desks;
//	HBITMAP bmLayer_Select;
//	HBITMAP bmLayer_Windows;

	int TitlebarHeight;

	int IconSize;
	int MinIconSize; // if bScaleToFit and icon scales below this, don't draw it
	RECT IconBorder; // keep icon out of this area

	bool bLoadInBox;          // delay LocalInit until we get a box hook
	bool bStayChildWindow;    // tells SetOnTop to never turn the window into an unparented popup
	bool bScaleChanged;       // force all windows to be recalculated and redrawn
	bool bGuessDeskNo;        // we're recovering from a crash, so use the guessed current desk number
	bool bGuessDeskCount;     // the user didn't specify the number of desks, so use the guess (unless it's 1, then use 2x2)
	bool bUnHideTaskOnSwitch; // tells MoveWindowVWM to show all windows

	bool bAlwaysOnTop;
	bool bAutoHide;
	bool bBackBevel;
	bool bBevel;
	bool bConstantUpdate;
	bool bDetectTitleChange;
//	bool bFixedSizeWindows;
//	bool bFocusCenter;  pretty arbitrary, ain't it?
	bool bFocusLast;
	bool bFocusTop;
	bool bForceIconBorderTop;
	bool bHideTaskOnSwitch;
	bool bNoBorder;
	bool bNoDrawSelBack;
	bool bNoGathering;
	bool bNoShellFlash;
	bool bNoSwitchOnFocus;
	bool bNoShow;
	bool bNoWindows;
	bool bNoWrap;
	bool bReturnToFirstScreen;
	bool bScaleIconToFit;
	bool bSelBevel;
	bool bShowIcons;
	bool bSnapWindowOnDrag;
	bool bSwitchOnAnyFocus;
	bool bTitlebars;
	bool bTitleForTitle;
	bool bTransparent;			// aka UseRealTrans
	bool bUseFakeTrans;
	bool bUseTransBlt;

	int SnapShotLevel;			// for miniwindows
	int SnapShotLevelVWM;		// for background
	int SnapShotLevelSel;		// for select
	HBITMAP bmBackSnapShot;
	HBITMAP bmSelSnapShot;

	s_EdgeMap emBack;       // main window background
	RECT rDeskBorder;       // amount to remove from the main window to get the desk area
	s_EdgeMap emSel;        // desk select background
	RECT rSelExpand;        // amount to add to the desk are to get the select rect
	s_EdgeMap emWnd;        // miniwindows
	s_EdgeMap emTitle;      // titlebars
	s_EdgeMap emFlash;      // flashing miniwindows
	s_EdgeMap emFlashTitle; // flashing miniwindow title

	int MatchDesk;          // desk to find windows on for wndspecs
	DWORD AtParams[10];     // arbitrary @var specific parameters (never actually set)

	const char *ConfigFile;

	char EscapeDollar;

	s_Action *OnSwitchFrom;
	s_Action *OnSwitchTo;
	s_Action *OnHover;
	s_Action *OnUnhover;
	s_Action *OnClickLeft;
	s_Action *OnClickRight;
	s_Action *OnClickMiddle;
	s_Action *OnMouseOver;
	s_Action *OnMouseOut;
  s_Action *OnMouseMove;
	s_Action *OnNewWindow;
	s_Action *OnRemoveWindow;
	s_Action *OnTitleChange;
	s_Action *OnInit;
	s_Action *OnQuit;
#ifdef INCLUDE_SCROLL_EVENTS
	s_Action *OnScrollUp;
	s_Action *OnScrollDown;
#endif
	s_Action *OnMoveWindow;
	s_Action *OnSizeWindow;
#ifdef INCLUDE_DROP_EVENTS
/* Now, the problem here is that there's no way to get such
   events with the old (WM_DROPFILES) method, and no guarantee
	 that you'll have a filename to make this useful with the
	 new (IDropTarget) method.  (pesky ID Lists)
*/
	s_Action *OnDropLeft;
	s_Action *OnDropRight;
	s_Action *OnDropMiddle;
#endif
	s_Action *OnFlash;
  s_Action *OnRepaint;

	RECT ParentRect;
//	SIZE PerDeskSize;

	int nSetupPass; // which phase of the few initial passes through UpdateWindowList we're on
  int InitialDesk;

#ifdef VWMDROPTARGET
	VWMDropTarget *pDropTarget;
#endif

	bool bItsEaster;
};

#pragma pack(push,4)
// structure size must be a multiple of four for use with StringEnum

struct rcbooldef {
	const char *key;
	bool s_GlobalData::*pItem;
	bool bDynamicChange;
};

#define DEFBOOL(n,f) { #n, &s_GlobalData::b##n, f },

struct rcactiondef {
	const char *key;
	s_Action *s_GlobalData::*pAction;
};

#define DEFACTION(n) { #n, &s_GlobalData::On##n },
#pragma pack(pop)

#define RCPREFIX "VWM"
#define RCPREFIXLEN 3
/*
struct rcitemdef {
	const char *key;
	void s_GlobalData::*pItem;
	DWORD def[4];
	void (*pReader)(rcitem *);
	void (*pUnreader)(rcitem *);
	void (*pParser)(rcitem *,const char *);
	bool bDynamicChange;
};
*/

#define BANGPREFIX "!VWM"
#define BANGPREFIXLEN 4

/* debug macros
	These will all compile to nothing in release builds except
	for LASTERROR, which evaluates to its exp parameter.

	ASSERT(cond)
		Prints a message to the logs on failure.
	ASSERTR(cond,value)
		Use this when continuing would be catastrophic.
		returns with value if assert fails, use ; for void functions
	DCLEAR(pointer)
		Clears pointers to NULL, only in debug build.
		Use for pointers that are not used as flags, only need to be
		NULL for error checking.

	TRACE(level,message)
		Add a message to the logs.  Note that level is currently a string
		used to filter the logs with CSS.
	CTRACE(level,cond,message)
		Add a message to the logs if cond evaluates to true.
	TRACELAST(exp)
		TRACELAST will evaluate to its exp only  in release builds.
		This will log the string associated with GetLastError if the
		exp returns 0.
	WTRACE(level,wnd,message)
		Add a message to the logs and associated window data.
		wnd should be a pointer to a s_WinData
	CWTRACE(level,cond,wnd,message)
		Conditionally add a message with window data.
	INITTRACE(level,message)
		Add a message to the logs during the initial set-up pass.
*/
#undef ASSERT
#ifndef NDEBUG
extern void AssertFail(const char *,const char *,int);
#define ASSERT(b)    if (!(b))   AssertFail(#b,__FILE__,__LINE__)
#define ASSERTR(b,r) if (!(b)) { AssertFail(#b,__FILE__,__LINE__); return r; }
#define DCLEAR(a)    a = NULL
extern bool bInLocalInit;
#ifndef WRITEABLE
#define WRITEABLE(ptr,len) (ptr!=NULL && !IsBadWritePtr(ptr,len))
#define READABLE(ptr,len) (ptr!=NULL && !IsBadReadPtr(ptr,len))
#define EXECABLE(ptr) (ptr!=NULL && !IsBadCodePtr(reinterpret_cast<int (__stdcall*)(void)>(ptr)))
#endif
#else
#define ASSERT(b) (void)0
#define ASSERTR(b,r) (void)0
#define DCLEAR(a) (void)0
#endif

// trace
#if !defined(NDEBUG) && defined(TRACE_FILE)
void Trace(const char *level,const char *line);

// startup: nothing is wrong, but display separately
#define TRACE_STARTUP "startup"
// status: nothing is wrong
#define TRACE_STATUS "status"
#define TRACE_BANG "status bang"
#define TRACE_DUMP "status dump"
// error: something is wrong in the code
#define TRACE_ERROR "error"
#define TRACE_ASSERT "error assert"
// warning: something is wrong in the environment
#define TRACE_WARNING "warning"

void TraceLastError(const char *,const char *,int);
void DumpWindowList(void);
void DumpStarLists(void);
void DumpWindow(const char *level,const s_WinData *pData);
void DumpBools(void);
void DumpActions(void);
#ifdef VWM_STATS
void DumpStats(void);
#endif
#define TRACE(level,a) Trace(level,a)
#define LASTERROR(n) if (n==0) TraceLastError(#n,__FILE__,__LINE__)
#define INITTRACE(level,a) if (bInLocalInit) TRACE(level,a)
#define CTRACE(level,c,a) if (c) TRACE(level,a)
#define TRACEBANG(a) { char b[256]; sprintf(b,"%s %s",a,args?args:"NULL"); TRACE(TRACE_BANG,b); }
#define WTRACE(level,w,a)  { char b[256]; sprintf(b,"%s: %08X/%10d",a,w?w->hWnd:0,w?w->hWnd:0); TRACE(level,b); DUMPWIN(level,w); }
#define CWTRACE(level,c,w,a) if (c) WTRACE(level,w,a)
#define DUMPWIN(level,a) DumpWindow(level,a)
#define DI int i=-1
#define CI i &=
#define CI0 i=-1
#else
#define TRACE(level,a) (void)0
#define LASTERROR(n) n
#define INITTRACE(level,a) (void)0
#define CTRACE(level,c,a) (void)0
#define TRACEBANG(a) (void)0
#define WTRACE(level,w,a) (void)0
#define CWTRACE(level,c,w,a) (void)0
#define DUMPWIN(level,a) (void)0
#define DI (void)0
#define CI
#define CI0 (void)0
#endif

// initial maximum, the array grows if needed
#define MAX_WINDOWS 50

// size of buffer for local storage across PaintWnd routines (on stack)
#define PAINTWND_BUFFER_SIZE 1024

// constants
#define MOUSE_ACTION_COUNT 3
// if the high bit is set, it's a special window type and the other flags aren't used
#define MATCH_WND_TITLE						0x00000001
#define MATCH_WND_CLASS						0x00000002
#define MATCH_WND_PAINTORDER			0x00010000
#define MATCH_WND_DESKONLY				0x00020000
#define MATCH_WND_ALL							0x00040000
#define MATCH_WND_SPECIAL					0x10000000
#define MATCH_WND_BYHANDLE				0x10000001
#define TRANSPARENT_COLOR RGB(255,0,255)

extern const s_PaintWnd Paint_Colors;
extern const s_PaintWnd Paint_Bitmap;
extern const s_PaintWnd Paint_SnapShot;
extern const s_MouseAction MouseTable[MOUSE_ACTION_COUNT];

extern void (WINAPI *SwitchToThisWindow)(HWND, int);
extern BOOL (WINAPI *IsHungAppWindow) (HWND);
extern BOOL (WINAPI *IsHungThread) (DWORD);
extern BOOL  (WINAPI *dAlphaBlend)( IN HDC, IN int, IN int, IN int, IN int, IN HDC, IN int, IN int, IN int, IN int, IN BLENDFUNCTION);
extern HWND (WINAPI *dGetAncestor)(HWND, UINT); // Not in win95, need to emulate (but only partially)
extern BOOL (WINAPI *dEnumDisplayMonitors)(IN HDC,IN LPCRECT,IN MONITORENUMPROC,IN LPARAM);
extern BOOL (WINAPI *dUpdateLayeredWindow)(HWND,HDC,POINT*,SIZE*,HDC,POINT*,COLORREF,BLENDFUNCTION*,DWORD);
// there's a check as to whether or not this is defined
#define dEnumDisplayMonitors dEnumDisplayMonitors
extern BOOL (WINAPI *dGetMonitorInfo)( IN HMONITOR, OUT LPMONITORINFO );
#define dGetMonitorInfo dGetMonitorInfo
extern HMONITOR (WINAPI *dMonitorFromPoint)(IN POINT,IN DWORD );
extern LSAPI int (*dParseCoordinate)(char strVal[], int defaultVal, int maxVal);

extern const char szAppName[];
extern const char szAppVer[];
extern const char szAppAuth[];

#ifndef WS_EX_LAYERED
#define WS_EX_LAYERED           0x00080000
#define ULW_COLORKEY            0x00000001
#define ULW_ALPHA               0x00000002
#define ULW_OPAQUE              0x00000004
#endif

// GetWindowLong macros

#define IsLiteStepWnd(hWnd) \
	(GetWindowLong(hWnd, GWL_USERDATA) == magicDWord)
#define IsToolWindow(hWnd) \
	((GetWindowLong(hWnd, GWL_EXSTYLE) & WS_EX_TOOLWINDOW) == WS_EX_TOOLWINDOW)
#define HasTitleBar(hWnd) \
	((GetWindowLong(hWnd, GWL_STYLE) & WS_CAPTION) == WS_CAPTION)
#define IsTopMost(hWnd) \
	((GetWindowLong(hWnd, GWL_EXSTYLE) & WS_EX_TOPMOST) == WS_EX_TOPMOST)

// global data
extern s_GlobalData GlobalData;
extern s_ParseData *g_pParseData;
extern rcbooldef RCBools[];
extern rcactiondef RCActions[];

// global functions

#define IsKeyDown(key) (GetKeyState(key)&0x8000)
#define ControlKeyDown() IsKeyDown(VK_CONTROL)
#define ShiftKeyDown() IsKeyDown(VK_SHIFT)
#define MouseButtonDown() \
 (IsKeyDown(VK_LBUTTON)		\
||IsKeyDown(VK_RBUTTON)		\
||IsKeyDown(VK_MBUTTON))

// RabidVWM.cpp
extern "C" {
__declspec( dllexport ) int initModuleEx(HWND parent, HINSTANCE dll, LPCSTR szPath);
__declspec( dllexport ) int quitModule(HINSTANCE dll);
__declspec( dllexport ) int initWharfModule(HWND ParentWnd, HINSTANCE dllInst, wharfDataType *wd);
__declspec( dllexport ) int quitWharfModule(HINSTANCE dllInst);
}
void SetOnTop(void); // update "ontop-ness" of the main window
void Refresh(void); // reread config, don't lose windows
int LocalInit(void);
int LocalQuit(void);
void AllocateWindows(int nSpace);
// colors.cpp
COLORREF ColorScale(COLORREF source,int delta);
void CalcBevelColors(s_BevelColors *,int amount);
COLORREF AddTrans(COLORREF c, int amount);
// windows.cpp
bool IsWndMinimized(const s_WinData *pData);
int FindKnownWindow(HWND hWnd);
s_WinData *GetWindowPtr(HWND hWnd);
void UpdateRect(s_WinData *pData);
void UpdateRectN(s_WinData *pData);
void UpdateWindowList(void); // bring the KnownWindows list up to date
bool MoveWindowVWM(s_WinData *pData, const SIZE *amount, int desk, int wndesk, HDWP hDwp);
bool MoveAllWindows(const SIZE *amount,int desk);
DWORD ParseWindowSpecFlags(const char *spec);
void GetWindowProc(s_WinData *pData,DWORD ptr);
void ExecWindowBySpec(const char *spec, WinModifier *func, DWORD lParam);
bool MoveWindowToDesk(s_WinData *pData,int nDesk,HDWP hDwp);
bool Gather(int desk);
bool GatherLost(int desk);
s_WinData *WindowFromVWMPoint(const POINT *pt);
bool ShouldDrawWnd(const s_WinData *pData);
bool ShouldMoveWnd(const s_WinData *pData);
bool ShouldHideWnd(const s_WinData *pData);
void SetVWMPos(void);
// desks.cpp
void GetDeskOff(int desk,POINT *off); // offset to top-left corner of desk in current state
void GetDeskDif(int olddesk,int newdesk,SIZE *dif); // independent of current state
void GetDeskDifXY(int oldx,int oldy,int newx,int newy,SIZE *dif);
bool SwitchDesk(int newdesk); // will return true on succuss, incl switch to current desk
int GetDeskFromDelta(int x,int y);
int DeskFromPoint(int x,int y);
void DeskFromPointXY(int x,int y,POINT *desk);
int DeskFromVWMPoint(int x,int y);
bool OutsideAnyDesk(const POINT *pt);
bool IsOnScreen(const RECT *area);
void ScreenToVWM(RECT *dest,const RECT *source);
void VWMToScreen(int x,int y,POINT *pt);
void GetDeskVWMRect(int desk, RECT *r);
void SetHomeDesk(s_WinData *pData,int home);
void GuessDesksFromWindows(POINT *dmin,POINT *dmax);
void SetDeskCount(int x,int y);
void SetDeskFocus(void);
void Resize(int width, int height);
// wndproc.cpp
LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
// readconfig.cpp
int ReadConfig(void);
int ReleaseConfig(void);
int MaxXCoord(void);
int MaxYCoord(void);
int ParseCoord(const char *source,int,int);
int ParseDimension(const char *source,int,int);
// paint.cpp
void Repaint(void);
void PaintVWM_Generic(HDC dc,void (*PaintDesk)(HDC,s_DeskData *,RECT *));
// wndmatchers.cpp
void ReadStarLists(void);
void UnreadStarLists(void);
void DoWindowMatches(s_WinData *pData, const char *title);
// bangs.cpp
void RegisterBangs(void);
void RemoveBangs(void);
int GetBangDeskNo(LPCSTR args, const char **next);
// star-flashmsg.cpp
void RealFlashMessageProc(CWPSTRUCT *msg);
// blitters.cpp
COLORREF GetPixelStd(HDC src, COLORREF back, int x, int y);
COLORREF GetPixelTrans(HDC src, COLORREF back, int x, int y);
void FastStretchBlt(
	HDC dest,
	int dw, int dh,
	HDC src,
	int sw, int sh,
	int method,
	GetPixelProc *GetPixel);
void UpdateDesktopBits(void);
#ifdef LSTOOLS_REV
#define TransEdgeBlt(a,b,c,d) LSTransEdgeBlt(a,b,c,d,GlobalData.bUseTransBlt)
#define StringEnum LSStringEnum
#define FreeDesktopBmp() (void)0
#else // no LSTOOLS_REV
BOOL TransEdgeBlt(
	HDC d, const RECT *rDest,
	HDC s, s_EdgeMap *emSrc);
void FreeDesktopBmp(void);
// parsers.cpp
int StringEnum(const char *string,const char **list,int inc=1);
#endif // LSTOOLS_REV
// atvars.cpp
void AtVarExec(s_Action *pAction,bool bGetMousePos,bool bWindowFromPoint);
// easter.cpp
void StartEaster(void);
void StopEaster(void);
void StepEaster(void);
void ShowEaster(PixelBlock *pdest);

