/*
  RabidVWM LiteStep Module - version 1.0
	Bang Commands

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "RabidVWM.h"
#include "multimonitor.h"
#include <stdlib.h>

#ifndef LSTOOLS_REV
struct bangcmddef {
	const char *Name;
	BangCommand *Command;
};
#else
typedef lsBangCmdDef bangcmddef;
#endif

void BangDesk(HWND caller, LPCSTR args);
void BangGather(HWND caller, LPCSTR args);
void BangGatherLost(HWND caller, LPCSTR args);
void BangToggleOnTop(HWND caller, LPCSTR args);
void BangOpen(HWND caller, LPCSTR args);
void BangShow(HWND caller, LPCSTR args);
void BangHide(HWND caller, LPCSTR args);
void BangMoveApp(HWND caller, LPCSTR args);
void BangReset(HWND caller, LPCSTR args);
void BangMove(HWND caller, LPCSTR args);
void BangFlash(HWND caller, LPCSTR args);
void BangExec(HWND caller, LPCSTR args);
void BangLoadBoxHook(HWND caller, LPCSTR args); // for lsbox
void BangSetDeskCount(HWND caller, LPCSTR args);
void BangToggle(HWND caller, LPCSTR args);
void BangSetAction(HWND caller, LPCSTR args);
void BangResize(HWND caller, LPCSTR args);
void BangToggleHideTask(HWND caller, LPCSTR args);
void BangRefreshBack(HWND caller, LPCSTR args);
void BangSticky(HWND caller, LPCSTR args);

// kept for lsdistro compatibility
void BangLeft(HWND caller, LPCSTR args);
void BangUp(HWND caller, LPCSTR args);
void BangRight(HWND caller, LPCSTR args);
void BangDown(HWND caller, LPCSTR args);

bangcmddef BangList[] = {
	{ "DESK",							&BangDesk							},
	{ "EXEC",							&BangExec							},
	{	"FLASH",						&BangFlash						},
	{ "GATHER",						&BangGather						},
	{ "GATHERLOST",				&BangGatherLost				},
	{ "HIDE",							&BangHide							},
	{ "LOADBOXHOOK",			&BangLoadBoxHook			}, // for lsbox
	{ "MOVE",							&BangMove							},
	{ "MOVEAPP",					&BangMoveApp					},
	{ "OPEN",							&BangOpen							},
	{ "REFRESHBACK",			&BangRefreshBack			},
	{	"RESET",						&BangReset						},
	{ "SIZE",				  		&BangResize						},	
	{ "SETDESKS",					&BangSetDeskCount			},
	{ "ACTION",		    		&BangSetAction				},
	{	"SHOW",							&BangShow							},
	{ "STICKY",						&BangSticky						},
	{ "TOGGLE",						&BangToggle						},
	{ "TOGGLEHIDETASK",		&BangToggleHideTask		},
	{ "TOGGLEONTOP",			&BangToggleOnTop			},

// kept for lsdistro compatibility
	{ "LEFT",							&BangLeft							},
	{ "UP",								&BangUp								},
	{ "RIGHT",						&BangRight						},
	{ "DOWN",							&BangDown							},
	{ NULL,								NULL									}
};

const char *ToggleStrings[] = {
	"TOGGLE", "YES", "TRUE", "ON", NULL
};

const char *SideStrings[] = {
	"LEFT", "TOP", "RIGHT", "BOTTOM", NULL
};

const char *DirectionStrings[] = {
	"LEFT", "UP", "RIGHT", "DOWN", NULL
};

const char *ShowStrings[] = {
	"TOGGLE", "HIDE", "OFF", "NO", "FALSE", NULL
};

#ifndef LSTOOLS_REV
void RegisterBangs(void)
{
	bangcmddef *pBang;
	char BangString[64] = BANGPREFIX;
	TRACE(TRACE_STATUS,"Registerring bang commands.");

	pBang = BangList;
	while (pBang->Name != NULL) {
		lstrcpy(BangString+BANGPREFIXLEN,pBang->Name);
		AddBangCommand(BangString, pBang->Command);
		TRACE(TRACE_STATUS,BangString);
		pBang++;
	}
}

void RemoveBangs(void)
{
	bangcmddef *pBang;
	char BangString[64] = BANGPREFIX;
	TRACE(TRACE_STATUS,"Removing bang commands.");

	pBang = BangList;
	while (pBang->Name != NULL) {
		lstrcpy(BangString+BANGPREFIXLEN,pBang->Name);
		RemoveBangCommand(BangString);
		TRACE(TRACE_STATUS,BangString);
		pBang++;
	}
}
#else
void RegisterBangs(void)
{ LSRegisterBangList(BANGPREFIX,BANGPREFIXLEN,BangList); }
void RemoveBangs(void)
{ LSRemoveBangList(BANGPREFIX,BANGPREFIXLEN,BangList); }
#endif

void GetBangDeskNoProc(s_WinData *pData,DWORD pDesk)
{ *((int *)pDesk) = pData->HomeDesk; }

int GetBangDeskNo(LPCSTR args, const char **next)
{
	char one[256],two[256];
	char *tokens[] = { one, two };
	int x,y;

	if (args==NULL) return GlobalData.CurrentDesk;

	if (next) *next = args;

	if (*args == '(') {
		/* get desk number from windowspec */ 
		const char *p1,*p2;
		char *p3;

		p1 = args+1;
		p2 = strchr(p1,')');
		p3 = one;

		if (p2) {
			/* copy out the windowspec */ 
			while (p1<p2) *p3++ = *p1++;
			/* terminate it            */ 
			*p3 = 0; x = -1;
			/* set next                */ 
			if (next) {
				*next = p2+1;
				while (**next==' ' || **next=='\t') ++*next;
			}
			/* get desk                */ 
			ExecWindowBySpec(one,&GetBangDeskNoProc,(DWORD)&x);


		#if !defined(NDEBUG) && defined(TRACE_FILE)
			char buf[256];
			sprintf(buf,"desk from spec was %d",x);
			TRACE(TRACE_STATUS,buf);
		#endif
		} else {
			/* starts with (, but no ) present */ 
			TRACE(TRACE_WARNING,"Missing ending paren for windowspec.");
			x = GlobalData.CurrentDesk;
		}
	} else {

		*one = *two = 0;
		CommandTokenize( args, tokens, 2, NULL);
		y = atoi(two);

		x = StringEnum(one,DirectionStrings);
		if (x!=-1) {
			/* skip first argument */ 
			if (next) {
				*next += lstrlen(one);
				/* now move past whitespace */ 
				while (**next==' ' || **next=='\t') ++*next;
			}

			if (y==0 && *two!='0') {
				y=1;
			} else {
				/* skip second argument */ 
				if (next) {
					*next += lstrlen(two);
					/* now move past whitespace */ 
					while (**next==' ' || **next=='\t') ++*next;
				}
			}

		#if !defined(NDEBUG) && defined(TRACE_FILE)
			char buf[256];
			sprintf(buf,"relative desk delta = %d",y);
			TRACE(TRACE_STATUS,buf);
		#endif

			switch (x) {
				case 0: /* left  */ 
					x = GetDeskFromDelta(-y, 0); break;
				case 1: /* up    */ 
					x = GetDeskFromDelta( 0,-y); break;
				case 2: /* right */ 
					x = GetDeskFromDelta( y, 0); break;
				case 3: /* down  */ 
					x = GetDeskFromDelta( 0, y); break;
				default:
					TRACE(TRACE_WARNING,"invalid keyword");
					break;
			}
		} else {
			TRACE(TRACE_STATUS,"Reading boring old desk #");

			x = atoi(one);

			if (x==0 && *one!='0') {
				x = GlobalData.CurrentDesk;
			} else {
				/* skip first argument */ 
				if (next) {
					*next += lstrlen(one);
					/* now move past whitespace */ 
					while (**next==' ' || **next=='\t') ++*next;
				}

				--x;

				if (!(y==0 && *two!='0')) {
					/* skip second argument */ 
					if (next) {
						*next += lstrlen(two);
						/* now move past whitespace */ 
						while (**next==' ' || **next=='\t') ++*next;
					}

					--y;
					x += y * GlobalData.nDesksX;
				}
			}
		}
	}

#if !defined(NDEBUG) && defined(TRACE_FILE)
	char buf[256];
	sprintf(buf,"returning desk #%d",x);
	TRACE(TRACE_STATUS,buf);
#endif

	if (x<0) x = GlobalData.CurrentDesk;
	return x;
}

// START OF BANG COMMANDS ***********************************************************

void BangDesk(HWND caller, LPCSTR args)
{
	TRACEBANG("!VWMDesk");
	if (GlobalData.Desks==NULL)
	{
		TRACE(TRACE_WARNING,"Called before desks were set up.");
		return;
	}
	SwitchDesk(GetBangDeskNo(args,NULL));
}

// kept for lsdistro compatibility (!VWMDesk is *much* more flexible)
void BangLeft (HWND caller, LPCSTR args) { SwitchDesk(GetDeskFromDelta(-1, 0)); }
void BangUp   (HWND caller, LPCSTR args) { SwitchDesk(GetDeskFromDelta( 0,-1)); }
void BangRight(HWND caller, LPCSTR args) { SwitchDesk(GetDeskFromDelta( 1, 0)); }
void BangDown (HWND caller, LPCSTR args) { SwitchDesk(GetDeskFromDelta( 0, 1)); }

void BangGather(HWND caller, LPCSTR args)
{
	TRACEBANG("!VWMGather");
	Gather(GetBangDeskNo(args,NULL));
}

void BangGatherLost(HWND caller, LPCSTR args)
{
	TRACEBANG("!VWMGatherLost");
	GatherLost(GetBangDeskNo(args,NULL));
}

void BangToggleOnTop(HWND caller, LPCSTR args)
{
	int i = 0;
	TRACEBANG("!VWMToggleOnTop");
	
	if (args && *args) i = StringEnum(args,ToggleStrings);

	if (i==0) GlobalData.bAlwaysOnTop = !GlobalData.bAlwaysOnTop;
	else if (i==-1) GlobalData.bAlwaysOnTop = false;
	else GlobalData.bAlwaysOnTop = true;

	SetOnTop();
}

void BangToggleHideTask(HWND caller, LPCSTR args)
{
	int i = 0;
	TRACEBANG("!VWMToggleHideTask");
	
	if (args && *args) i = StringEnum(args,ToggleStrings);

	if (i==0) GlobalData.bHideTaskOnSwitch = !GlobalData.bHideTaskOnSwitch;
	else if (i==-1) GlobalData.bHideTaskOnSwitch = false;
	else GlobalData.bHideTaskOnSwitch = true;

	/* activate changes */ 
	GlobalData.bUnHideTaskOnSwitch = !GlobalData.bHideTaskOnSwitch;
	SwitchDesk(GlobalData.CurrentDesk);
	GlobalData.bUnHideTaskOnSwitch = false;
}

// toggle most available bools from rc
void BangToggle(HWND caller, LPCSTR args)
{
	rcbooldef *pDef;
	char one[MAX_LINE_LENGTH],two[MAX_LINE_LENGTH];
	char *tokens[] = { one, two };
	int i,j;

	TRACEBANG("!VWMToggle");

	*one = *two = 0;
	LCTokenize(args, tokens, 2, NULL);

	/* nothing to do */ 
	if (!*one) return;

	i = 0;
	if (*two) i = StringEnum(two,ToggleStrings);

	j = StringEnum(one,(const char **)RCBools,sizeof(rcbooldef)/sizeof(const char **));
	if (j>=0) {
		pDef = RCBools + j;
		if (pDef->bDynamicChange) {
			if (i==0) GlobalData.*pDef->pItem = !(GlobalData.*pDef->pItem);
			else if (i==-1) GlobalData.*pDef->pItem = false;
			else GlobalData.*pDef->pItem = true;
      TRACE(TRACE_STATUS,"bChangeState due to !VWMToggle");
			GlobalData.bChangeState = true;
		}
	}
}

// set event triggered action
// "!VWMSetAction OnAction !command"
// or "!VWMSetAction OnAction" to clear it
void BangSetAction(HWND caller, LPCSTR args)
{
	char one[MAX_LINE_LENGTH],two[MAX_LINE_LENGTH];
	char *tokens[] = { one };
	int nArgs,i;
	rcactiondef *pDef;

	TRACEBANG("!VWMSetAction");

	if (!args) return;

	*two = 0;
	nArgs = LCTokenize(args, tokens, 1, two);
	

	i = StringEnum(one,(const char **)RCActions,sizeof(rcactiondef)/sizeof(const char **));

	if (i>=0) {
		pDef = RCActions + i;
		/* clear it */ 
		if ((GlobalData.*pDef->pAction)!=NULL) {
			TRACE(TRACE_STATUS,"Clearing.");
			delete[] (void *)(GlobalData.*pDef->pAction)->Command;
			(GlobalData.*pDef->pAction)->Command = NULL;
		}
		/* set it   */ 
		if (*two) {
			TRACE(TRACE_STATUS,"Setting.");
      if ((GlobalData.*pDef->pAction)==NULL)
      {
        GlobalData.*pDef->pAction = new s_Action;
        ZeroMemory(GlobalData.*pDef->pAction,sizeof(s_Action));
      }
			(GlobalData.*pDef->pAction)->Command = new char[lstrlen(two)+1];
			lstrcpy((char *)(GlobalData.*pDef->pAction)->Command,two);
		} else {
      if ((GlobalData.*pDef->pAction)!=NULL)
      {
			  /* *really* clear it */ 
			  delete (GlobalData.*pDef->pAction);
			  (GlobalData.*pDef->pAction) = NULL; /* used as flag */ 
      }
		}
	}
}

void BangOpen(HWND caller, LPCSTR args)
{
	int desk;
	const char *command;
	TRACEBANG("!VWMOpen");

	desk = GetBangDeskNo(args,&command);

	if (desk>=0 && desk<GlobalData.nDesks) SwitchDesk(desk);

	LSExecute(NULL, command, SW_SHOWNORMAL);
}

void BangExec(HWND caller, LPCSTR args)
{
	TRACEBANG("!VWMExec");
	s_Action act;
	act.Command = args;
	act.Hold = 0;
	AtVarExec(&act,false,false);
}

void BangShow(HWND caller, LPCSTR args)
{
	int i = StringEnum(args,ShowStrings);
	TRACEBANG("!VWMShow");

	if (i==-1)
		GlobalData.bNoShow = false;
	else if (i==0)
		GlobalData.bNoShow = !GlobalData.bNoShow;
	else
		GlobalData.bNoShow = true;
	ShowWindow(GlobalData.hMainWnd,GlobalData.bNoShow?SW_HIDE:SW_SHOWNA);
}

void BangHide(HWND caller, LPCSTR args)
{
	TRACEBANG("!VWMHide");
	GlobalData.bNoShow = true;
	ShowWindow(GlobalData.hMainWnd,SW_HIDE);
}

void BangMoveAppProc(s_WinData *pData,DWORD desk)
{ MoveWindowToDesk(pData,desk,NULL); }

void BangMoveApp(HWND caller, LPCSTR args)
{
	int desk;
	const char *windowspec;
	TRACEBANG("!VWMMoveApp");

	desk = GetBangDeskNo(args,&windowspec);

	ExecWindowBySpec(windowspec,&BangMoveAppProc,(DWORD)desk);
}

void BangReset(HWND caller, LPCSTR args)
{
	TRACEBANG("!VWMReset");
	Refresh();
}

void BangFlashProc(s_WinData *pData,DWORD state)
{
	switch ((int)state)
	{
		case 0:  pData->bFlash = !pData->bFlash; break;
		case -1: pData->bFlash = false; break;
		default: pData->bFlash = true;  break;
	}
	pData->bFlashOn = false;
	pData->iFlashCountLeft = GlobalData.iFlashCount;
}

void BangFlash(HWND caller, LPCSTR args)
{
	char one[MAX_LINE_LENGTH],two[MAX_LINE_LENGTH];
	char *tokens[] = { one };
	int i;

	TRACEBANG("!VWMFlash");

	*one = *two = 0;
	if (*args=='#')
	{
		/* toggle strings don't start with #, so this must be windowspec */ 
		lstrcpy(two,args);
	}
	else
	{
		LCTokenize(args, tokens, 1, two);
	}
	/* two = window spec         */ 
	/* one = toggle string       */ 

	i = 0;
	if (*one) i = StringEnum(one,ToggleStrings);

	ExecWindowBySpec(two,&BangFlashProc,i);
}

void BangStickyProc(s_WinData *pData,DWORD state)
{
	bool b;
	switch ((int)state)
	{
		case 0:  b = !pData->bSticky; break;
		case -1: b = false; break;
		default: b = true;  break;
	}
	/* make sure it's on the current desk */ 
	if (b)
	{
		MoveWindowToDesk(pData,GlobalData.CurrentDesk,NULL);
	}
	else
	{
		pData->bPosChanged = true;
		UpdateRect(pData);
	}
	pData->bSticky = b;
}

void BangSticky(HWND caller, LPCSTR args)
{
	char one[MAX_LINE_LENGTH],two[MAX_LINE_LENGTH];
	char *tokens[] = { one };
	int i;

	TRACEBANG("!VWMSticky");

	*one = *two = 0;
	if (*args=='#')
	{
		/* toggle strings don't start with #, so this must be windowspec */ 
		lstrcpy(two,args);
	}
	else
	{
		LCTokenize(args, tokens, 1, two);
	}
	/* two = window spec         */ 
	/* one = toggle string       */ 

	/* if *two==0, last focused window is used */ 
	/* (in ExecWindowBySpec)                   */ 

	i = 0;
	if (*one) i = StringEnum(one,ToggleStrings);

	ExecWindowBySpec(two,&BangStickyProc,i);

	Repaint();
}

// load this module into lsbox
// !VWMLoadBoxHook "blah" "blah" "blah" ... path\to.box hwnd
void BangLoadBoxHook(HWND caller, LPCSTR args)
{
	char *p;
	const char *rest;
	int i;
  char one[MAX_LINE_LENGTH],two[MAX_LINE_LENGTH],three[MAX_LINE_LENGTH];
  char *tokens[] = { one, two, three };

	rest = args;
	if (args && *args)
	{
    *one = *two = *three = 0;
		while (GetToken(rest,tokens[0],&rest,FALSE))
		{
			p = tokens[0];
			tokens[0] = tokens[1];
			tokens[1] = tokens[2];
			tokens[2] = p;
		}
		/* tokens[2] = window handle */ 
		i = atoi(tokens[2]);
    if (!i)
    {
      sprintf(tokens[1],"Invalid !VWMLoadBoxHook: %s%s.","bad HWND = ",tokens[2]);
			MessageBox(NULL,tokens[1],szAppName,MB_OK|MB_ICONWARNING);
      return;
    }
    GlobalData.hParentWnd = (HWND)i;
		/* tokens[1] = box file */ 
		if (*tokens[1]!=0)
		{
			if (GlobalData.ConfigFile) delete[] (char *)GlobalData.ConfigFile;
			/* copy it                                    */ 
			GlobalData.ConfigFile = new char[lstrlen(tokens[1])+1];
			/* l.. functions are winapi rather than crt   */ 
			lstrcpy((char *)GlobalData.ConfigFile,tokens[1]);
			TRACE(TRACE_STATUS,GlobalData.ConfigFile);
		}

		if (!GlobalData.bLoadInBox) {
			TRACE(TRACE_WARNING,"Stuffed in a box without warning.");
			/* not a good way to do things, ask them to warn me next time */ 
			MessageBox(NULL,
				"Please add VWMLoadInBox to your LiteStep configuration file.",
				szAppName,MB_OK|MB_ICONWARNING);
			TRACE(TRACE_STATUS,"Closing to reopen in box.");
			/* now destroy the main window and recreate it in the box     */ 
			LocalQuit();
			GlobalData.bLoadInBox = true;

			TRACE(TRACE_STATUS,"Now reopenning...");
		}
		/* we will be a child window always */ 
		/* (this disables always on top)    */ 
		GlobalData.bStayChildWindow = true;
		/* create main window in box        */ 
		LocalInit();
		return;
  }
  sprintf(tokens[1],"Invalid !VWMLoadBoxHook: %s%s.","no arguments","");
	MessageBox(NULL,tokens[1],szAppName,MB_OK|MB_ICONWARNING);
}

void BangSetDeskCount(HWND caller, LPCSTR args)
{
	int x,y,deskx,desky;
	char one[MAX_LINE_LENGTH],two[MAX_LINE_LENGTH];
	char *tokens[] = { one, two };

	TRACEBANG("!VWMSetDesks");

	*one = *two = 0;
	LCTokenize(args, tokens, 2, NULL);

	x = y = 0;
	if (*one) {
		if (*one=='-' || *one=='+' || *one=='0') x = GlobalData.nDesksX;
		x += atoi(one);
	}
	if (*two) {
		if (*two=='-' || *two=='+' || *two=='0') y = GlobalData.nDesksY;
		y += atoi(two);
	}

	deskx = GlobalData.CurrentDesk%GlobalData.nDesksX;
	desky = GlobalData.CurrentDesk/GlobalData.nDesksX;
	SetDeskCount(x,y);
	deskx += GlobalData.nDesksX*desky;
	GlobalData.CurrentDesk = deskx;
}

void BangMove(HWND caller, LPCSTR args)
{
	char one[256],two[256],three[256];
	char *tokens[] = { one,two };
	int count,x,y;
	TRACEBANG("!VWMMove");

	if (strchr("0123456789%-",*args))
	{	
		count = CommandTokenize(args,tokens,2,NULL);
	}
	else
	{
		count = CommandTokenize(args,tokens,1,three);
		LSMONITOR m = LSGetMonitor(one);

		/* if it's not a valid monitor, ignore it */ 
		if (m!=LSMONITOR_NONE)
			LSGetMonitorScreenRect(m,&GlobalData.rcScreen);

		count = CommandTokenize(three,tokens,2,NULL);
	}
	if (count==2) {
		x = dParseCoordinate(one,GlobalData.VWMPos.x-GlobalData.rcScreen.left,
					GlobalData.rcScreen.right-GlobalData.rcScreen.left)
			+ GlobalData.rcScreen.left;
		y = dParseCoordinate(two,GlobalData.VWMPos.y-GlobalData.rcScreen.top,
					GlobalData.rcScreen.bottom-GlobalData.rcScreen.top)
			+ GlobalData.rcScreen.top;
	} else {
		x = GlobalData.VWMPos.x;
		y = GlobalData.VWMPos.y;
	}
	GlobalData.VWMPos.x = x;
	GlobalData.VWMPos.y = y;

  SetVWMPos();

  BangRefreshBack(caller,args);
}

void BangResize(HWND caller, LPCSTR args)
{
	char one[MAX_LINE_LENGTH],two[MAX_LINE_LENGTH];
	char *tokens[] = { one, two };
	int w,h;

	TRACEBANG("!VWMResize");

	CommandTokenize(args, tokens, 2, NULL);

	w = dParseCoordinate(one,GlobalData.VWMSize.cx,GlobalData.rcScreen.right-GlobalData.rcScreen.left);
	h = dParseCoordinate(two,GlobalData.VWMSize.cy,GlobalData.rcScreen.bottom-GlobalData.rcScreen.top);

	Resize(w,h);
}

void BangRefreshBack(HWND caller, LPCSTR args)
{
	UpdateDesktopBits();
	Repaint();
}


