/*
  RabidVWM LiteStep Module - version 1.0
	General Bitmap Functions

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "RabidVWM.h"
#include "multimonitor.h"
#include <stdlib.h> // for atoi in PaintDesktopEx

#define TRACE_DUMP_DESKTOPBMP TRACE_DUMP " deskbmp"
#define TRACE_WARN_DESKTOPBMP TRACE_WARNING " deskbmp"
#define TRACE_ERROR_DESKTOPBMP TRACE_ERROR " deskbmp"
#define TRACE_STATUS_DESKTOPBMP TRACE_STATUS " deskbmp"

// default is TransparentBltLS, which is VERY SLOW
//#define USE_TRANSBLT
// TransparentBlt, which I can't get to link
//#define USE_BITBLT
// BitBlt, which doesn't let you have transparency, but it's fast
// AlphaBlend might be another idea, but it's probly pretty slow too.

typedef COLORREF PointSampler(HDC, COLORREF back, CONST POINT *, CONST POINT *, GetPixelProc *);

// PointSamplers
COLORREF Sample1Pt(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel);
COLORREF Sample5Plus(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel);
COLORREF Sample5X(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel);
COLORREF Sample9Octagon(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel);
COLORREF Sample8Octagon(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel);
COLORREF Sample17Insane(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel);
#define POINT_SAMPLERS 6

PointSampler *SamplerTable[POINT_SAMPLERS] = {
	&Sample1Pt,
	&Sample5Plus,
	&Sample5X,
	&Sample9Octagon,
	&Sample8Octagon,
	&Sample17Insane,
};

COLORREF SampleNull(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel)
{ return (COLORREF)src; }


/* used to get snapshots MUCH faster than StretchBlt            */ 
/* StretchBlt seems to run over the entire source coords always */ 
/*		both start at 0                                           */ 
/*		source>dest                                               */ 
void FastStretchBlt(
	HDC dest,
	int dw, int dh,
	HDC src,
	int sw, int sh,
	int method,
	GetPixelProc *GetPixel)
{
	int x,y;
	POINT step,rem,here;
	COLORREF c;
	PointSampler *Sample;

	CTRACE(TRACE_ERROR,sw<dw,"FSBing to larger bitmap.");
	CTRACE(TRACE_ERROR,sh<dh,"FSBing to larger bitmap.");

	if (method<0 || method>=POINT_SAMPLERS)
		Sample = &SampleNull;
	else
		Sample = SamplerTable[method]; 

	step.x = sw/(dw+1)-1;
	step.y = sh/(dh+1)-1;
	rem.y = sh%dh;

	here.y = sh - step.y - 1;
	for (y=dh-1;y>=0;--y) {
		here.x = sw - step.x - 1;
		rem.x = sw%dw;
		for (x=dw-1;x>=0;--x) {
			c = ::GetPixel(dest,x,y);
			c = Sample(src,c,&here,&step,GetPixel);
			SetPixelV(dest, x, y, c);

			here.x -= step.x;
			if (rem.x--) here.x--;
		}
		here.y -= step.y;
		if (rem.y--) here.y--;
	}
}

COLORREF GetPixelStd(HDC src, COLORREF back, int x, int y)
{
	if (!PtVisible(src,x,y)) return back;
	return ::GetPixel(src,x,y);
}

COLORREF GetPixelTrans(HDC src, COLORREF back, int x, int y)
{
	COLORREF c;
	c = GetPixelStd(src,back,x,y);
	if (c==TRANSPARENT_COLOR) return back;
	return c;
}
/* not currently used
COLORREF GetPixelAlpha(HDC src, COLORREF back, int x, int y)
{
	COLORREF c;
	BYTE alpha;
	c = ::GetPixel(src,x,y);
	alpha = c>>(8*3);
	if (alpha) {
		if (alpha!=255) {
			int r,g,b;
			r = MulDiv(GetRValue(c),255-alpha,255) + MulDiv(GetRValue(back),alpha,255);
			g = MulDiv(GetGValue(c),255-alpha,255) + MulDiv(GetGValue(back),alpha,255);
			b = MulDiv(GetBValue(c),255-alpha,255) + MulDiv(GetBValue(back),alpha,255);
			c = RGB(r,g,b);
		} else {
			// fully transparent
			c = back
		}
	}
	// alpha component has been removed
	return c;
}
*/

COLORREF Sample1Pt(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel)
{ return GetPixel(src,back,pt->x,pt->y); }

COLORREF xSample4Plus(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel)
{
	COLORREF c;
	c  = (GetPixel(src,back,pt->x-step->x/2,pt->y          )&0xfcfcfcfc)>>2;
	c += (GetPixel(src,back,pt->x          ,pt->y-step->y/2)&0xfcfcfcfc)>>2;
	c += (GetPixel(src,back,pt->x+step->x/2,pt->y          )&0xfcfcfcfc)>>2;
	c += (GetPixel(src,back,pt->x          ,pt->y+step->y/2)&0xfcfcfcfc)>>2;
	return c;
}

/* rounding error = 1 */ 
COLORREF Sample5Plus(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel)
{
	COLORREF c;
	c  = (GetPixel(src,back,pt->x          ,pt->y          )&0xfefefefe)>>1;
	c += (xSample4Plus(src,back,pt,step,GetPixel)           &0xfefefefe)>>1;
	return c;
}

COLORREF xSample4X(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel)
{
	COLORREF c;
	c  = (GetPixel(src,back,pt->x-step->x*7/20,pt->y-step->y*7/20)&0xf8f8f8f8)>>2;
	c += (GetPixel(src,back,pt->x-step->x*7/20,pt->y+step->y*7/20)&0xf8f8f8f8)>>2;
	c += (GetPixel(src,back,pt->x+step->x*7/20,pt->y-step->y*7/20)&0xf8f8f8f8)>>2;
	c += (GetPixel(src,back,pt->x+step->x*7/20,pt->y+step->y*7/20)&0xf8f8f8f8)>>2;
	return c;
}

/* rounding error = 1 */ 
COLORREF Sample5X(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel)
{
	COLORREF c;
	c  = (GetPixel(src,back,pt->x             ,pt->y             )&0xfefefefe)>>1;
	c += (xSample4X(src,back,pt,step,GetPixel)										&0xfefefefe)>>1;
	return c;
}

/* rounding error = 1 */ 
COLORREF Sample9Octagon(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel)
{
	COLORREF c;
	c  = (GetPixel(src,back,pt->x             ,pt->y             )&0xfefefefe)>>1;
	c += (xSample4X(src,back,pt,step,GetPixel)										&0xfcfcfcfc)>>2;
	c += (xSample4Plus(src,back,pt,step,GetPixel)									&0xfcfcfcfc)>>2;
	return c;
}

/* rounding error = 2, compensated */ 
COLORREF Sample8Octagon(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel)
{
	COLORREF c;
	c  =  GetPixel(src,back,pt->x             ,pt->y             )&0x03030303;
	c += (xSample4X(src,back,pt,step,GetPixel)										&0xfefefefe)>>1;
	c += (xSample4Plus(src,back,pt,step,GetPixel)									&0xfefefefe)>>1;
	return c;
}

COLORREF xSample8Octagon22_5(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel)
{
	/* octagon, rotated 22.5 degrees */ 
	/* distance about 2*step         */ 
	COLORREF c;
	c  = (GetPixel(src,back,pt->x-step->x*9/10,pt->y-step->y*4/10)&0xf8f8f8f8)>>3;
	c += (GetPixel(src,back,pt->x-step->x*4/10,pt->y+step->y*9/10)&0xf8f8f8f8)>>3;
	c += (GetPixel(src,back,pt->x+step->x*9/10,pt->y-step->y*4/10)&0xf8f8f8f8)>>3;
	c += (GetPixel(src,back,pt->x+step->x*4/10,pt->y+step->y*9/10)&0xf8f8f8f8)>>3;
	c += (GetPixel(src,back,pt->x-step->x*4/10,pt->y-step->y*9/10)&0xf8f8f8f8)>>3;
	c += (GetPixel(src,back,pt->x-step->x*9/10,pt->y+step->y*4/10)&0xf8f8f8f8)>>3;
	c += (GetPixel(src,back,pt->x+step->x*4/10,pt->y-step->y*9/10)&0xf8f8f8f8)>>3;
	c += (GetPixel(src,back,pt->x+step->x*9/10,pt->y+step->y*4/10)&0xf8f8f8f8)>>3;
	return c;
}

/* rounding error = 16, compensated */ 
COLORREF Sample17Insane(HDC src,COLORREF back,CONST POINT *pt,CONST POINT *step, GetPixelProc *GetPixel)
{
	COLORREF c;
	/* rounding error is fixed in subroutines? */ 
	c  = (xSample4X          (src,back,pt,step,GetPixel)&0xfefefefe)>>1;
	c += (xSample4Plus       (src,back,pt,step,GetPixel)&0xfcfcfcfc)>>2;
	c += (xSample8Octagon22_5(src,back,pt,step,GetPixel)&0xfcfcfcfc)>>2;
	return c;
}

#ifndef LSTOOLS_REV

#define min3(a,b,c) (((a)<(b))?(min(a,c)):(min(b,c)))

// tile a piece of the source over a piece of the destination
// with sl,st aligned to x,y of destination
BOOL TileBltAligned(
	HDC d, int dl, int dt, int dw, int dh,
	HDC s, int sl, int st, int sw, int sh,
	int x, int y,
	DWORD rop)
{
	int i,j,top;
	BOOL ret = TRUE;

	top = -sh-(dt-y)%sh;
	i   = -sw-(dl-x)%sw;

	for (i;i<dw;i+=sw) {
		for (j=top;j<dh;j+=sh) {
			ret &=
			BitBlt(
				d,max(dl,dl+i),max(dt,dt+j),
				  min3(sw,dw-i,sw+i),min3(sh,dh-j,sh+j),
				s,max(sl,sl-i),max(st,st-j),rop);
		}
	}
	CTRACE(TRACE_WARNING,!ret,"TileBlt: failed");
	return ret;
}

// tile a piece of the source over a piece of the destination
BOOL TileBlt(
	HDC d, int dl, int dt, int dw, int dh,
	HDC s, int sl, int st, int sw, int sh,
	DWORD rop)
{
#if(1)
	return TileBltAligned(d,dl,dt,dw,dh,s,sl,st,sw,sh,dl,dt,rop);
#else
	int i,j;
	BOOL ret = TRUE;
	for (i=0;i<dw;i+=sw) {
		for (j=0;j<dh;j+=sh) {
			ret &=
			BitBlt(
				d,dl+i,dt+j,
				min(sw,dw-i),min(sh,dh-j),
				s,sl,st,rop);
		}
	}
	CTRACE(TRACE_WARNING,!ret,"TileBlt: failed");
	return ret;
#endif
}

#ifndef NDEBUG
static int teb_lev = 0;
#endif

BOOL TransEdgeBlt(
	HDC d, const RECT *rDest,
	HDC s, s_EdgeMap *emSrc)
{
	/* from my own scbar/RabidBitmap                      */ 
	/* varnames have been changed to protect the innocent */ 
	/* This could be done in one pass with 9 blts, but    */ 
	/* I believe there were problems with StretchBlt and  */ 
	/* bitmaps that were only a few pixels wide and       */ 
	/* stretched vertically.  It was always off by one.   */ 
	struct { int w,h; } center;
	struct { int l,t,r,b; } squish;
	int bw,bh;
	DI;
#ifdef VWM_STATS
	int start = GetTickCount();
#endif

	const int dl = rDest->left, dt = rDest->top;
	const int dw = rDest->right-rDest->left, dh = rDest->bottom-rDest->top;
	const int sl = 0, st = 0;
	const int sw = emSrc->Size.cx, sh = emSrc->Size.cy;
	const int bl = emSrc->Edge.left, bt = emSrc->Edge.top;
	const int br = emSrc->Edge.right, bb = emSrc->Edge.bottom;
	const bool tile = emSrc->Tile;

	HDC dc1,dc2;
	HBITMAP tempbm1,tempbm2;
	HBITMAP old1,old2;

	BOOL ret = TRUE;

	center.w = dw - bl - br;
	center.h = dh - bt - bb;

	bw = sw - bl - br;
	bh = sh - bt - bb;

  /* fix for missing center when edges add up to image size */ 
  if (bw==0) bw=1;
  if (bh==0) bh=1;

	/* stretch horizontally */ 
	if (dw == sw) {
		/* skip the first stretch step */ 
		dc1 = s;
	} else {
		/* create temp dc & bitmap     */ 
		dc1 = CreateCompatibleDC(d); StatGDIAdd(DC);
		tempbm1 = CreateCompatibleBitmap(d,dw,sh); StatGDIAdd(Bitmap);
		old1 = (HBITMAP)SelectObject( dc1, tempbm1 );

		CTRACE(TRACE_WARNING,!tempbm1,"TEB: Failed to create temp bitmap 1.");

		if (center.w < 0) {
			/* squeeze borders           */ 
			squish.l = dw*bl/(bl+br);
			squish.r = dw-squish.l;
			center.w = 0;
		} else {
			/* stretch/squeeze center    */ 
			squish.l = bl;
			squish.r = br;
		}
		if (squish.l)
			ret &= BitBlt(
				dc1,0,0,squish.l,sh,
				s,0,0,
				SRCCOPY);
		if (center.w) {
			if (tile)
				ret &= TileBlt(
					dc1,squish.l,0,center.w,sh,
					s,squish.l,0,bw,sh,
					SRCCOPY);
			else
				ret &= StretchBlt(
					dc1,squish.l,0,center.w,sh,
					s,squish.l,0,bw,sh,
					SRCCOPY);
		}
		if (squish.r)
			ret &= BitBlt(
				dc1,dw-squish.r,0,squish.r,sh,
				s,sw-squish.r,0,
				SRCCOPY);

		CTRACE(TRACE_WARNING,!ret,"TEB: First pass failed.");
	}

	/* stretch vertically */ 
	if (dh == sh) {
		/* skip second stretch step */ 
		dc2 = dc1;
	} else {
		/* create temp dc & bitmap  */ 
		dc2 = CreateCompatibleDC(d); StatGDIAdd(DC);
		tempbm2 = CreateCompatibleBitmap(d,dw,dh); StatGDIAdd(Bitmap);
		old2 = (HBITMAP)SelectObject( dc2, tempbm2 );

		CTRACE(TRACE_WARNING,!tempbm2,"TEB: Failed to create temp bitmap 2.");

		if (center.h < 0) {
			/* squeeze borders        */ 
			squish.t = dh*bt/(bt+bb);
			squish.b = dh-squish.t;
			center.h = 0;
		} else {
			/* stretch/squeeze center */ 
			squish.t = bt;
			squish.b = bb;
		}
		if (squish.t)
			ret &= BitBlt(
				dc2,0,0,dw,squish.t,
				dc1,0,0,
				SRCCOPY);
		if (center.h) {
			if (tile)
				ret &= TileBlt(
					dc2,0,squish.t,dw,center.h,
					dc1,0,squish.t,dw,bh,
					SRCCOPY);
			else
				ret &= StretchBlt(
					dc2,0,squish.t,dw,center.h,
					dc1,0,squish.t,dw,bh,
					SRCCOPY);
		}
		if (squish.b)
			ret &= BitBlt(
				dc2,0,dh-squish.b,dw,squish.b,
				dc1,0,sh-squish.b,
				SRCCOPY);

		CTRACE(TRACE_WARNING,!ret,"TEB: Second pass failed.");
	}

	/* draw overlay */ 
	if (GlobalData.bUseTransBlt)
		TransparentBltLS(
			d,dl,dt,dw,dh,
			dc2,0,0,
			TRANSPARENT_COLOR);
	else
		BitBlt(
			d,dl,dt,dw,dh,
			dc2,0,0,
			SRCCOPY);

	/* free things  */ 
	if (dc1 != s) {
		SelectObject(dc1, old1);
		CI DeleteObject(tempbm1); StatGDIDel(Bitmap);
		CI DeleteDC(dc1); StatGDIDel(DC);
	}
	if (dc2 != dc1) {
		SelectObject(dc2, old2);
		CI DeleteObject(tempbm2); StatGDIDel(Bitmap);
		CI DeleteDC(dc2); StatGDIDel(DC);
	}

	CTRACE(TRACE_WARNING,!i,"TEB: Failed to delete stuff.");

#ifdef VWM_STATS
	VWMStats.EdgeBltTime += GetTickCount()-start;
#endif

	return ret;
}


//
//  Desktop Bitmap Things
//


HDC hdcDesktop = 0;
HBITMAP hbmDesktop = 0;

static void GetSnapShot_Desktop(HBITMAP &bm,int level,int width,int height)
{
	TRACE(TRACE_STATUS_DESKTOPBMP,"GetSnapShot_Desktop");

	if (hdcDesktop) {
		HDC bdc = CreateCompatibleDC(hdcDesktop); StatGDIAdd(DC);
		CTRACE(TRACE_WARN_DESKTOPBMP,!bdc,"Failed to create bdc");
		HBITMAP hbm = CreateCompatibleBitmap(hdcDesktop,width,height); StatGDIAdd(Bitmap);
		CTRACE(TRACE_WARN_DESKTOPBMP,!hbm,"Failed to create hbm");
		HBITMAP bOld = (HBITMAP)SelectObject(bdc,hbm);

		if (bdc && hbm) {
			FastStretchBlt(
				bdc,width,height,
				hdcDesktop,GlobalData.ScreenSize.cx,GlobalData.ScreenSize.cy,
				level-1,&GetPixelStd);
			if (bm) DeleteObject(bm);
			bm = hbm; hbm = NULL;
		}

		SelectObject(bdc,bOld);
		if (hbm)
		{
			DeleteObject(hbm);
			StatGDIDel(Bitmap);
		}
		DeleteDC(bdc);
		StatGDIDel(DC);
	}
	else
	{
		TRACE(TRACE_WARN_DESKTOPBMP,"No hdcDesktop to get snapshot from");
	}
}

// PaintDesktopEx and RegQueryStringValue were stolen from Maduin's label module.
// This version has been mutilated to fit my needs, so any bugs are my fault.

LPTSTR RegQueryStringValue(HKEY hKey, LPCTSTR pszSubKey, LPCTSTR pszValueName, LPTSTR pszBuffer, int nBufferSize, LPTSTR pszDefaultValue);

bool GetSmallestScreenSizeProc(LSMONITOR Mon,LPARAM Size)
{
	RECT rScreen;

	LSGetMonitorScreenRect(Mon,&rScreen);

	if (rScreen.right-rScreen.left<reinterpret_cast<SIZE *>(Size)->cx
		&&rScreen.bottom-rScreen.top<reinterpret_cast<SIZE *>(Size)->cy)
	{
		reinterpret_cast<SIZE *>(Size)->cx = rScreen.right-rScreen.left;
		reinterpret_cast<SIZE *>(Size)->cy = rScreen.bottom-rScreen.top;
	}

	return true;
}

struct PDCData
{
	HDC hdcDesktop;
	HDC hdcWallpaper;
	SIZE sWallpaper;
	RECT rDesktop;
};

bool PaintDesktopCenterredProc(LSMONITOR Mon,LPARAM Data)
{
	RECT rScreen;
	RECT rSource; /* right&bottom are width&height */ 
	POINT pos;

	LSGetMonitorScreenRect(Mon,&rScreen);

	/* alight top-left monitor to 0,0 of bitmap */ 
	OffsetRect(&rScreen,
		-reinterpret_cast<PDCData *>(Data)->rDesktop.left,
		-reinterpret_cast<PDCData *>(Data)->rDesktop.top);

	pos.x = (rScreen.right+rScreen.left - reinterpret_cast<PDCData *>(Data)->sWallpaper.cx)/2;
	pos.y = (rScreen.bottom+rScreen.top - reinterpret_cast<PDCData *>(Data)->sWallpaper.cy)/2;

	/* check size in x direction */ 
	if (pos.x<rScreen.left)
	{
		/* crop it to the size of the screen */ 
		rSource.left = rScreen.left-pos.x;
		rSource.right = rScreen.right-rScreen.left;
		pos.x = rScreen.left;
	}
	else
	{
		/* no need to crop it */ 
		rSource.left = 0;
		rSource.right = reinterpret_cast<PDCData *>(Data)->sWallpaper.cx;
	}

	/* check size in y direction */ 
	if (pos.y<rScreen.top)
	{
		/* crop it to the size of the screen */ 
		rSource.top = rScreen.top-pos.y;
		rSource.bottom = rScreen.bottom-rScreen.top;
		pos.y = rScreen.top;
	}
	else
	{
		/* no need to crop it */ 
		rSource.top = 0;
		rSource.bottom = reinterpret_cast<PDCData *>(Data)->sWallpaper.cy;
	}

	/* paint it */ 
	BitBlt(
		reinterpret_cast<PDCData *>(Data)->hdcDesktop,
		pos.x,pos.y,
		rSource.right,rSource.bottom,
		reinterpret_cast<PDCData *>(Data)->hdcWallpaper,
		rSource.left,rSource.top,
		SRCCOPY);

	return true;
}

/*
Load the background bitmap and paint it the way Windows would.
Some of these seem pretty brain dead, but there's nothing I can
do about it.

This takes a noticable amount of time, especially with larger
desktop area, so don't call it very often at all.

On return phdc will have a device context with the bitmap selected
into it, and phbm will have the bitmap to select into the device
contect to get the desktop bitmap back out.
*/
void LoadDesktopBitmap(HDC *phdc,HBITMAP *phbm)
{
	PDCData d;
	SIZE sScreen;

	ASSERT(phdc!=NULL);
	ASSERT(phbm!=NULL);

	HBITMAP hbmWallpaper;
	HBITMAP hbmDesktop;
	HDC hdcScreen;

	hdcScreen = GetDC(NULL);
	CTRACE(TRACE_WARN_DESKTOPBMP,!hdcScreen,"Failed to get screen DC");

	d.hdcDesktop = CreateCompatibleDC(hdcScreen);
	StatGDIAdd(DC);
	CTRACE(TRACE_WARN_DESKTOPBMP,!d.hdcDesktop,"Failed to create hdcDesktop");

	if (!LSGetMonitorScreenRect(LSMONITOR_ALL,&d.rDesktop))
		TRACE(TRACE_WARN_DESKTOPBMP,"Failed: LSGetMonitorScreenRect(LSMONITOR_ALL)");
	int cxDesktop = d.rDesktop.right-d.rDesktop.left;
	int cyDesktop = d.rDesktop.bottom-d.rDesktop.top;

	hbmDesktop = CreateCompatibleBitmap(hdcScreen, cxDesktop, cyDesktop);
	StatGDIAdd(Bitmap);
	CTRACE(TRACE_WARN_DESKTOPBMP,!hbmDesktop,"Failed to create hbmDesktop");
	hbmDesktop = (HBITMAP) SelectObject(d.hdcDesktop, hbmDesktop);

	RECT r;
	SetRect(&r, 0, 0, cxDesktop, cyDesktop);
	FillRect(d.hdcDesktop, &r, GetSysColorBrush(COLOR_DESKTOP));

	char wallpaper[MAX_PATH] = { 0 };

	RegQueryStringValue(HKEY_CURRENT_USER,
		"Control Panel\\Desktop", "Wallpaper", wallpaper,
		MAX_PATH, "");
	TRACE(TRACE_DUMP_DESKTOPBMP,"Wallpaper =");
	TRACE(TRACE_DUMP_DESKTOPBMP,wallpaper);

	if(wallpaper[0])
	{
		int wallpaperStyle;
		int tileWallpaper;

		char buffer[32];

		RegQueryStringValue(HKEY_CURRENT_USER, "Control Panel\\Desktop", "WallpaperStyle", buffer, 32, "0");
		wallpaperStyle = atoi(buffer);
		TRACE(TRACE_DUMP_DESKTOPBMP,"WallpaperStyle =");
		TRACE(TRACE_DUMP_DESKTOPBMP,buffer);

		RegQueryStringValue(HKEY_CURRENT_USER, "Control Panel\\Desktop", "TileWallpaper", buffer, 32, "0");
		tileWallpaper = atoi(buffer);
		TRACE(TRACE_DUMP_DESKTOPBMP,"TileWallpaper = ");
		TRACE(TRACE_DUMP_DESKTOPBMP,buffer);

		hbmWallpaper = (HBITMAP) LoadImage(0, wallpaper, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
		StatGDIAdd(Bitmap);
		CTRACE(TRACE_WARN_DESKTOPBMP,!hbmWallpaper,"Failed to load wallpaper");

		if(hbmWallpaper)
		{
			BITMAP bm;
			GetObject(hbmWallpaper, sizeof(BITMAP), &bm);

			d.sWallpaper.cx = bm.bmWidth;
			d.sWallpaper.cy = bm.bmHeight;

			d.hdcWallpaper = CreateCompatibleDC(hdcScreen);
			CTRACE(TRACE_WARN_DESKTOPBMP,!d.hdcWallpaper,"Failed to create hdcScreen");
			StatGDIAdd(DC);
			hbmWallpaper = (HBITMAP) SelectObject(d.hdcWallpaper, hbmWallpaper);

			if(tileWallpaper)
			{
/* Tile behavior for multiple monitors:

		Bitmap is tiled across all monitors,
		aligned to upper-left of primary (ie, 0,0)
*/
				TRACE(TRACE_STATUS_DESKTOPBMP,"Tiling");

				TileBltAligned(d.hdcDesktop,
					0, 0,
					cxDesktop, cyDesktop,
					d.hdcWallpaper,
					0, 0,
					d.sWallpaper.cx, d.sWallpaper.cy,
					-d.rDesktop.left,-d.rDesktop.top,
					SRCCOPY);
			}
			else
			{
				if(wallpaperStyle == 2) /* stretch */ 
				{
/* Stretch behavior for multiple monitors:

		Bitmap is stretched/shrunk to fit smallest
		monitor then centered to each.
*/
					TRACE(TRACE_STATUS_DESKTOPBMP,"Stretching");

					sScreen.cx = sScreen.cy = 2147483647L;
					LSForEachMonitor(&GetSmallestScreenSizeProc,reinterpret_cast<LPARAM>(&sScreen));

					/* should always be true, but just in case */ 
					if (sScreen.cx<2147483647L)
					{
						HDC hdcStretch = CreateCompatibleDC(hdcScreen);
						CTRACE(TRACE_WARN_DESKTOPBMP,!hdcStretch,"Failed to create hdcStretch");
						HBITMAP hbmStretch = CreateCompatibleBitmap(hdcScreen,sScreen.cx,sScreen.cy);
						CTRACE(TRACE_WARN_DESKTOPBMP,!hbmStretch,"Failed to create hbmStretch");
						hbmStretch = (HBITMAP)SelectObject(hdcStretch,hbmStretch);

						SetStretchBltMode(hdcDesktop, STRETCH_DELETESCANS);

						StretchBlt(hdcStretch,
							0, 0,
							sScreen.cx, sScreen.cy,
							d.hdcWallpaper,
							0, 0,
							d.sWallpaper.cx, d.sWallpaper.cy,
							SRCCOPY);

						/* exchange it with the wallpaper */ 
						hbmWallpaper = (HBITMAP) SelectObject(d.hdcWallpaper, hbmWallpaper);
						DeleteDC(d.hdcWallpaper);
						DeleteObject(hbmWallpaper);

						d.hdcWallpaper = hdcStretch;
						hbmWallpaper = hbmStretch;
						d.sWallpaper = sScreen;

						/* then fall into the center code */ 
					}
					else
					{
						TRACE(TRACE_ERROR_DESKTOPBMP,"Failed: GetSmallestScreenSizeProc");
					}
				}

/* Center behavior for multiple monitors:

		Bitmap is cropped to each monitor, centered
		with edges that don't fit simply cropped off.
		(Did I mention there's cropping involved?)
*/
				TRACE(TRACE_STATUS_DESKTOPBMP,"Centering");

				LSForEachMonitor(&PaintDesktopCenterredProc,reinterpret_cast<LPARAM>(&d));
			}

			hbmWallpaper = (HBITMAP) SelectObject(d.hdcWallpaper, hbmWallpaper);
			DeleteDC(d.hdcWallpaper);
			StatGDIDel(DC);

			DeleteObject(hbmWallpaper);
			StatGDIDel(Bitmap);
		}
	}

	ReleaseDC(0, hdcScreen);

	*phdc = d.hdcDesktop;
	*phbm = hbmDesktop;
}

/*				UpdateDesktopBits

		Get the bits of the desktop that we need
		and cache them for later.  Get shrunken
		snapshots for back and sel, and the little
		bit that goes behind the vwm window for
		transparency.

		This needs to be called:
		 - on startup
		 - when the background changes
		 - when the vwm's position has changed
		 - when the vwm's size has changed
*/
void UpdateDesktopBits(void)
{
// need to be global for now for snapshot functions
//	HDC hdcDesktop;
//	HBITMAP hbmDesktop;

	/* don't waste time */ 
	if (!GlobalData.SnapShotLevelVWM
		&&!GlobalData.SnapShotLevelSel
		&&!GlobalData.bUseFakeTrans)
	{
		TRACE(TRACE_STATUS_DESKTOPBMP,"Nothing to use bg for, skipping");
		return;
	}

	LoadDesktopBitmap(&hdcDesktop,&hbmDesktop);

	/* shrunken snapshots */ 
	if (GlobalData.SnapShotLevelVWM)
		GetSnapShot_Desktop(GlobalData.bmBackSnapShot,GlobalData.SnapShotLevelVWM,
			GlobalData.VWMDeskSize.cx,GlobalData.VWMDeskSize.cy);
	if (GlobalData.SnapShotLevelSel)
		GetSnapShot_Desktop(GlobalData.bmSelSnapShot,GlobalData.SnapShotLevelSel,
			GlobalData.VWMDeskSize.cx+GlobalData.rSelExpand.left+GlobalData.rSelExpand.right,
			GlobalData.VWMDeskSize.cy+GlobalData.rSelExpand.top+GlobalData.rSelExpand.bottom);

	/* transparent bit */ 
	if (GlobalData.bUseFakeTrans)
	{
		HDC hdcTrans;
		HBITMAP hbm;

		TRACE(TRACE_STATUS_DESKTOPBMP,"Getting back for bUseFakeTrans");

		if (GlobalData.bmTransparent)
		{
			DeleteObject(GlobalData.bmTransparent);
			StatGDIDel(Bitmap);
			GlobalData.bmTransparent = NULL;
		}

		GlobalData.bmTransparent =
			CreateCompatibleBitmap(hdcDesktop,GlobalData.VWMSize.cx,GlobalData.VWMSize.cy);
		StatGDIAdd(Bitmap);
		CTRACE(TRACE_WARN_DESKTOPBMP,!GlobalData.bmTransparent,"Failed to create GlobalData.bmTransparent");

		hdcTrans = CreateCompatibleDC(hdcDesktop);
		hbm = (HBITMAP)SelectObject(hdcTrans,GlobalData.bmTransparent);
		CTRACE(TRACE_WARN_DESKTOPBMP,!hdcTrans,"Failed to create hdcTrans");

		RECT rDesktop;
		if (!LSGetMonitorScreenRect(LSMONITOR_ALL,&rDesktop))
		{
			TRACE(TRACE_ERROR_DESKTOPBMP,"Failed: LSGetMonitorScreenRect(LSMONITOR_ALL)");
		}

		BitBlt(hdcTrans, 0, 0, GlobalData.VWMSize.cx, GlobalData.VWMSize.cy,
			hdcDesktop, GlobalData.VWMPos.x-rDesktop.left, GlobalData.VWMPos.y-rDesktop.top,
			SRCCOPY);

		SelectObject(hdcTrans,hbm);
		DeleteDC(hdcTrans);
	}

	hbmDesktop = (HBITMAP)SelectObject(hdcDesktop,hbmDesktop);
	DeleteObject(hbmDesktop);
	StatGDIDel(Bitmap);
	DeleteDC(hdcDesktop);
	StatGDIDel(DC);
}

void FreeDesktopBmp(void)
{
	if (GlobalData.bmTransparent)
	{
		DeleteObject(GlobalData.bmTransparent);
		StatGDIDel(Bitmap);
		GlobalData.bmTransparent = NULL; // i think this counts as a flag
	}
}

#if(0)
// enhanced version of PaintDesktop
void PaintDesktopEx(HDC hdcDest, int xDest, int yDest, int cxDest, int cyDest, int xSrc, int ySrc, BOOL updateCache)
{
	if(hdcDesktop == 0 || updateCache)
	{
		if(!updateCache && hdcDest == 0)
			return;

		LoadDesktopBitmap(&hdcDesktop,&hbmDesktop);

		if (GlobalData.SnapShotLevelVWM)
			GetSnapShot_Desktop(GlobalData.bmBackSnapShot,GlobalData.SnapShotLevelVWM,
				GlobalData.VWMDeskSize.cx,GlobalData.VWMDeskSize.cy);
		if (GlobalData.SnapShotLevelSel)
			GetSnapShot_Desktop(GlobalData.bmSelSnapShot,GlobalData.SnapShotLevelSel,
				GlobalData.VWMDeskSize.cx+GlobalData.rSelExpand.left+GlobalData.rSelExpand.right,
				GlobalData.VWMDeskSize.cy+GlobalData.rSelExpand.top+GlobalData.rSelExpand.bottom);
	}

	if(hdcDest)
	{
		RECT rDesktop;
		LSGetMonitorScreenRect(LSMONITOR_ALL,&rDesktop);
		BitBlt(hdcDest, xDest, yDest, cxDest, cyDest,
			hdcDesktop, xSrc-rDesktop.left, ySrc-rDesktop.top, SRCCOPY);
	}

	if(hbmDesktop != 0)
	{
		hbmDesktop = (HBITMAP) SelectObject(hdcDesktop, hbmDesktop);
		DeleteObject(hbmDesktop);
		StatGDIDel(Bitmap);
		hbmDesktop = 0;
	}

	if(hdcDesktop != 0)
	{
		DeleteDC(hdcDesktop);
		StatGDIDel(DC);
		hdcDesktop = 0;
	}
}

void FreeDesktopBmp(void)
{
	if(hbmDesktop != 0)
	{
		hbmDesktop = (HBITMAP) SelectObject(hdcDesktop, hbmDesktop);
		DeleteObject(hbmDesktop);
		StatGDIDel(Bitmap);
		hbmDesktop = 0;
	}

	if(hdcDesktop != 0)
	{
		DeleteDC(hdcDesktop);
		StatGDIDel(DC);
	}
}
#endif

// seems a bit inefficient to do this for each string,
// but i'm lazy so I'll let Maduin's code do it.
// retrieve a string value from the registry
LPTSTR RegQueryStringValue(HKEY hKey, LPCTSTR pszSubKey, LPCTSTR pszValueName, LPTSTR pszBuffer, int nBufferSize, LPTSTR pszDefaultValue)
{
	DWORD dwDataType;
	DWORD dwDataSize = (DWORD) nBufferSize;
	HKEY hSubKey = NULL;
	LONG lErrorCode;
	
	if(pszSubKey)
	{
		lErrorCode = RegOpenKeyEx(hKey, pszSubKey, 0, KEY_QUERY_VALUE, &hSubKey);
		
		if(lErrorCode != ERROR_SUCCESS)
		{
			lstrcpyn(pszBuffer, pszDefaultValue, nBufferSize);
			return pszBuffer;
		}
		
		hKey = hSubKey;
	}
	
	lErrorCode = RegQueryValueEx(hKey, pszValueName, NULL, &dwDataType,
		(LPBYTE) pszBuffer, &dwDataSize);
	
	if(hSubKey)
		RegCloseKey(hSubKey);
		
	if(dwDataType != REG_SZ || lErrorCode != ERROR_SUCCESS)
		lstrcpyn(pszBuffer, pszDefaultValue, nBufferSize);
	
	return pszBuffer;
}
#else // LSTOOLS_REV
HDC hdcDesktop = 0;
HBITMAP hbmDesktop = 0;

static void GetSnapShot_Desktop(HBITMAP &bm,int level,int width,int height)
{
	TRACE(TRACE_STATUS_DESKTOPBMP,"GetSnapShot_Desktop");

	if (hdcDesktop) {
		HDC bdc = CreateCompatibleDC(hdcDesktop); StatGDIAdd(DC);
		CTRACE(TRACE_WARN_DESKTOPBMP,!bdc,"Failed to create bdc");
		HBITMAP hbm = CreateCompatibleBitmap(hdcDesktop,width,height); StatGDIAdd(Bitmap);
		CTRACE(TRACE_WARN_DESKTOPBMP,!hbm,"Failed to create hbm");
		HBITMAP bOld = (HBITMAP)SelectObject(bdc,hbm);

		if (bdc && hbm) {
			FastStretchBlt(
				bdc,width,height,
				hdcDesktop,GlobalData.ScreenSize.cx,GlobalData.ScreenSize.cy,
				level-1,&GetPixelStd);
			if (bm) DeleteObject(bm);
			bm = hbm; hbm = NULL;
		}

		SelectObject(bdc,bOld);
		if (hbm)
		{
			DeleteObject(hbm);
			StatGDIDel(Bitmap);
		}
		DeleteDC(bdc);
		StatGDIDel(DC);
	}
	else
	{
		TRACE(TRACE_WARN_DESKTOPBMP,"No hdcDesktop to get snapshot from");
	}
}

void UpdateDesktopBits(void)
{
// need to be global for now for snapshot functions
//	HDC hdcDesktop;
//	HBITMAP hbmDesktop;

	/* don't waste time */ 
	if (!GlobalData.SnapShotLevelVWM
		&&!GlobalData.SnapShotLevelSel
		&&!GlobalData.bUseFakeTrans)
		return;

	HDC hdcScreen = GetDC(0);
	hdcDesktop = CreateCompatibleDC(hdcScreen);
	RECT rDesktop;
	LSGetMonitorScreenRect(LSMONITOR_ALL,&rDesktop);
	int cxDesktop = rDesktop.right-rDesktop.left;
	int cyDesktop = rDesktop.bottom-rDesktop.top;
	hbmDesktop = CreateCompatibleBitmap(hdcScreen,cxDesktop,cyDesktop);
	ReleaseDC(0,hdcScreen);
	
	hbmDesktop = (HBITMAP)SelectObject(hdcDesktop,hbmDesktop);
	LSPaintDesktop(hdcDesktop,0,0,cxDesktop,cyDesktop,rDesktop.left,rDesktop.top,FALSE);

	/* shrunken snapshots */ 
	if (GlobalData.SnapShotLevelVWM)
		GetSnapShot_Desktop(GlobalData.bmBackSnapShot,GlobalData.SnapShotLevelVWM,
			GlobalData.VWMDeskSize.cx,GlobalData.VWMDeskSize.cy);
	if (GlobalData.SnapShotLevelSel)
		GetSnapShot_Desktop(GlobalData.bmSelSnapShot,GlobalData.SnapShotLevelSel,
			GlobalData.VWMDeskSize.cx+GlobalData.rSelExpand.left+GlobalData.rSelExpand.right,
			GlobalData.VWMDeskSize.cy+GlobalData.rSelExpand.top+GlobalData.rSelExpand.bottom);

	/* transparent bit */ 
	if (GlobalData.bUseFakeTrans)
	{
		HDC hdcTrans;
		HBITMAP hbm;

		if (GlobalData.bmTransparent)
		{
			DeleteObject(GlobalData.bmTransparent);
			StatGDIDel(Bitmap);
			GlobalData.bmTransparent = NULL;
		}

		GlobalData.bmTransparent =
			CreateCompatibleBitmap(hdcDesktop,GlobalData.VWMSize.cx,GlobalData.VWMSize.cy);
		StatGDIAdd(Bitmap);

		hdcTrans = CreateCompatibleDC(hdcDesktop);
		hbm = (HBITMAP)SelectObject(hdcTrans,GlobalData.bmTransparent);

		RECT rDesktop;
		LSGetMonitorScreenRect(LSMONITOR_ALL,&rDesktop);

		BitBlt(hdcTrans, 0, 0, GlobalData.VWMSize.cx, GlobalData.VWMSize.cy,
			hdcDesktop, GlobalData.VWMPos.x-rDesktop.left, GlobalData.VWMPos.y-rDesktop.top,
			SRCCOPY);

		SelectObject(hdcTrans,hbm);
		DeleteDC(hdcTrans);
	}

	hbmDesktop = (HBITMAP)SelectObject(hdcDesktop,hbmDesktop);
	DeleteObject(hbmDesktop);
	DeleteDC(hdcDesktop);
}
#endif