/*
  RabidVWM LiteStep Module - version 1.0
	Color Processing

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "RabidVWM.h"

// scale a color's intensity by delta
COLORREF ColorScale(COLORREF source,int delta)
{
	int intensity;
	int r,g,b;

	if (source&0xff000000) return source;

	r = GetRValue(source);
	g = GetGValue(source);
	b = GetBValue(source);

	intensity = 
		(r*30/100)+ // 30% red
		(g*59/100)+ // 59% green
		(b*11/100); // 11% blue

	r = r + max(intensity,0)*delta/255;
	g = g + max(intensity,0)*delta/255;
	b = b + max(intensity,0)*delta/255;
	if (r<0) r=0;
	if (g<0) g=0;
	if (b<0) b=0;

	intensity = 0;
	if (r>255) { intensity+=r-255; r=255; }
	if (g>255) { intensity+=g-255; g=255; }
	if (b>255) { intensity+=b-255; b=255; }
	r = min(r+intensity/2,255);
	g = min(g+intensity/2,255);
	b = min(b+intensity/2,255);
	
	return RGB(r,g,b);
}

void CalcBevelColors(s_BevelColors *c, int amount)
{
	ASSERT(c!=NULL);

	c->Light = ColorScale(c->Flat,amount);
	c->Dark  = ColorScale(c->Flat,-amount);
//	c->Black = ColorScale(c->Flat,-160);
//	c->White = ColorScale(c->Flat,160);
}

COLORREF AddTrans(COLORREF c, int amount)
{
	return RGB(
		amount*GetRValue(c)/255,
		amount*GetGValue(c)/255,
		amount*GetBValue(c)/255)
		+ 0x01000000*amount;
}

