/*
  RabidVWM LiteStep Module - version 1.0
	IDropTarget Implementation for Drag and Drop

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "RabidVWM.h"
#include <Shlobj.h>
#include <ShellAPI.h>

#include "droptarget.h"

#define USE_PIDL

#ifdef VWMDROPTARGET

void SetDropHover(void)
{ GlobalData.bInDragDrop = true; }
void ClearDropHover(void)
{ GlobalData.bInDragDrop = false; }

UINT VWMDropTarget::cf_types[] =
{
	0, /* CFSTR_SHELLIDLIST */ 
	CF_HDROP,
	0, /* CFSTR_INETURL			*/ 
	0, /* CFSTR_SHELLURL		*/ 
	CF_TEXT,
	CF_UNICODETEXT,
	0
};

/*

	CF_HDROP:
		works for regular files
		does not work for folders
		never sent for "special" things

	CFSTR_SHELLIDLIST:
		works for regular files
		works for folders, but complains
		works for "special" as for corresponding regular type
		does not work for control panel applets

*/

VWMDropTarget::DropFuncPtr VWMDropTarget::cf_actions[] =
{
	&VWMDropTarget::Drop_SHELLIDLIST,
	&VWMDropTarget::Drop_HDROP,
	&VWMDropTarget::Drop_INETURL,
	&VWMDropTarget::Drop_SHELLURL,
	&VWMDropTarget::Drop_TEXT,
	&VWMDropTarget::Drop_UNICODETEXT,
	0
};

void VWMDropTarget::RegisterFormats(void)
{
	// Why did Microsoft think it would be a good idea to force people to do this?
	cf_types[0] = RegisterClipboardFormat(CFSTR_SHELLIDLIST);
	cf_types[2] = RegisterClipboardFormat(CFSTR_INETURL);
	cf_types[3] = RegisterClipboardFormat(CFSTR_SHELLURL);
}


HRESULT STDMETHODCALLTYPE VWMDropTarget::QueryInterface( 
	/* [in] */ REFIID riid,
	/* [iid_is][out] */ void __RPC_FAR *__RPC_FAR *ppvObject)
{
	return E_NOINTERFACE;
}

ULONG STDMETHODCALLTYPE VWMDropTarget::AddRef( void)
{
	return 1;
}

ULONG STDMETHODCALLTYPE VWMDropTarget::Release( void)
{
	return 1;
}

HRESULT STDMETHODCALLTYPE VWMDropTarget::DragEnter( 
	/* [unique][in] */ IDataObject *pDataObj,
	/* [in] */ DWORD grfKeyState,
	/* [in] */ POINTL pt,
	/* [out][in] */ DWORD *pdwEffect)
{
	FORMATETC etc;
	ZeroMemory(&etc,sizeof(FORMATETC));

	TRACE(TRACE_STATUS,"DragEnter");

	SetDropHover();

	/* We want all of the content, in a global handle. */ 
	etc.tymed = TYMED_HGLOBAL;
	etc.dwAspect = DVASPECT_CONTENT;
	etc.lindex = -1;

	cf_index = 0;
	while (cf_types[cf_index]!=0)
	{
		etc.cfFormat = cf_types[cf_index];
		if (pDataObj->QueryGetData(&etc)==S_OK)
		{
			*pdwEffect = DROPEFFECT_COPY;
			return S_OK;
		}
		++cf_index;
	}
	*pdwEffect = DROPEFFECT_NONE;
	return S_OK;
}

HRESULT STDMETHODCALLTYPE VWMDropTarget::DragOver( 
	/* [in] */ DWORD grfKeyState,
	/* [in] */ POINTL pt,
	/* [out][in] */ DWORD *pdwEffect)
{
	SetDropHover();

	// forward it to the window to get hovering
	// mouse messages don't get sent to the main window during drag&drop
	SendMessage(GlobalData.hMainWnd,WM_MOUSEMOVE,(WPARAM)grfKeyState,MAKELONG(pt.x,pt.y));

	return S_OK;
}

HRESULT STDMETHODCALLTYPE VWMDropTarget::DragLeave( void)
{
	TRACE(TRACE_STATUS,"DragLeave");

	ClearDropHover();

	return S_OK;
}

HRESULT STDMETHODCALLTYPE VWMDropTarget::Drop( 
	/* [unique][in] */ IDataObject *pDataObj,
	/* [in] */ DWORD grfKeyState,
	/* [in] */ POINTL pt,
	/* [out][in] */ DWORD *pdwEffect)
{
	FORMATETC etc;
	STGMEDIUM stg;
	ZeroMemory(&etc,sizeof(FORMATETC));

	ClearDropHover();

	/* We want all of the content, in a global handle. */ 
	etc.tymed = TYMED_HGLOBAL;
	etc.dwAspect = DVASPECT_CONTENT;
	etc.lindex = -1;

	etc.cfFormat = cf_index;
	if (pDataObj->QueryGetData(&etc)!=S_OK)
	{
		cf_index = 0;
		while (cf_types[cf_index]!=0)
		{
			etc.cfFormat = cf_types[cf_index];
			if (pDataObj->QueryGetData(&etc)==S_OK)
				break;
			++cf_index;
		}
		// return if we can't process any data types provided
		if (cf_types[cf_index]==0)
		{
			TRACE(TRACE_STATUS,"Drop -- no matching data types");
			return S_OK;
		}
	}

	/* Grab the stuff */ 
	if (pDataObj->GetData(&etc,&stg)==S_OK)
	{
		/* switch to the right desktop */ 
		POINT p = { pt.x, pt.y };
		ScreenToClient(GlobalData.hMainWnd,&p);
		SwitchDesk(DeskFromVWMPoint(p.x,p.y));

		/* Use it, locked so we get a memory reference. */ 
		(this->*cf_actions[cf_index])(GlobalLock(stg.hGlobal));
		GlobalUnlock(stg.hGlobal);

		/* Free the medium. */ 
		if (stg.pUnkForRelease==NULL)
			GlobalFree(stg.hGlobal);
		else
			stg.pUnkForRelease->Release();
	}

	return S_OK;
}

#define HIDA_GetPIDLParent(pida) (LPCITEMIDLIST)(((LPBYTE)pida)+(pida)->aoffset[0])
#define HIDA_GetPIDLItem(pida, i) (LPCITEMIDLIST)(((LPBYTE)pida)+(pida)->aoffset[i+1])

#ifdef USE_PIDL

void DumpPIDL(LPCITEMIDLIST pidl)
{
#if !defined(NDEBUG) && defined(TRACE_FILE)
	char buffer[MAX_PATH];
	int j;
#endif
#if !defined(NDEBUG) && defined(TRACE_FILE)
		sprintf(buffer,"PIDL:%08X",pidl);
		TRACE(TRACE_DUMP,buffer);

		j = ((SHITEMID *)(pidl))->cb;
		buffer[j+j] = 0;
		while (j--)
		{
			/* convert to hex (scary lookin, innit?) */ 
			buffer[j+j+1] = (reinterpret_cast<const SHITEMID *>(pidl)->abID[j]&0x0f)["0123456789ABCDEF"];
			buffer[j+j+0] = (reinterpret_cast<const SHITEMID *>(pidl)->abID[j] >> 4)["0123456789ABCDEF"];
		}
		TRACE(TRACE_DUMP,buffer);

		j = (reinterpret_cast<const SHITEMID *>(pidl))->cb;
		buffer[j] = 0;
		while (j--)
		{
			buffer[j] = (reinterpret_cast<const SHITEMID *>(pidl)->abID[j]);
			if (buffer[j]<32 || buffer[j]>126) buffer[j] = '-';
		}
		TRACE(TRACE_DUMP,buffer);

		SHGetPathFromIDList(reinterpret_cast<LPCITEMIDLIST>(pidl),buffer);
		TRACE(TRACE_DUMP,buffer);
#endif
}

/*
void SetExplorerIfNecessary(SHELLEXECUTEINFO *lpse)
{
	SHFILEINFO sfi;

	ZeroMemory(&sfi,sizeof(sfi));
	SHGetFileInfo((LPCSTR)lpse->lpIDList,0,&sfi,sizeof(sfi),SHGFI_PIDL|SHGFI_ATTRIBUTES);
	if (sfi.dwAttributes&SFGAO_FOLDER)
	{
		lpse->fMask = SEE_MASK_IDLIST;
		TRACE("SetExplorerIfNecessary: using explorer");
	}
	else
	{
		lpse->fMask = SEE_MASK_IDLIST;
		TRACE("SetExplorerIfNecessary: not");
	}
}
*/

UINT GetSizePIDL(LPCITEMIDLIST pidl)
{
	UINT i = (pidl->mkid.cb);

	if (i==0) return sizeof(pidl->mkid.cb);

	TRACE(TRACE_DUMP,"** sizing");
	DumpPIDL(pidl);

	// walk the id list chain
	return i + GetSizePIDL(
		reinterpret_cast<LPCITEMIDLIST>(
		reinterpret_cast<const BYTE *>(pidl) + i
		));
}

LPITEMIDLIST AppendPIDL(LPMALLOC pMalloc,LPCITEMIDLIST pidlBase, LPCITEMIDLIST pidlAdd)
{
    if(pidlBase == NULL)
        return NULL;
    if(pidlAdd == NULL)
        return const_cast<LPITEMIDLIST>(pidlBase);
    
    LPITEMIDLIST pidlNew;

/* Microsoft has an example for this, but it only works
   if the id lists contain only one element each.
	 (either that, or they do exactly the same thing as
	 this, but don't tell you how to get GetSizePIDL)
*/

    UINT cb1 = GetSizePIDL(pidlBase);
    UINT cb2 = GetSizePIDL(pidlAdd);

    pidlNew = (LPITEMIDLIST)pMalloc->Alloc(cb1 + cb2);
    if (pidlNew)
    {
        CopyMemory(((LPSTR)pidlNew), pidlBase, cb1);
        CopyMemory(((LPSTR)pidlNew) + cb1 - sizeof(pidlBase->mkid.cb), pidlAdd, cb2);

				TRACE(TRACE_DUMP,"-- result");
				DumpPIDL(pidlNew);
    }
    return pidlNew;
}

#endif /* USE_PIDL */ 


void VWMDropTarget::Drop_SHELLIDLIST(void *files)
{
	SHELLEXECUTEINFO ExecInfo;
	CIDA *pCIDA = reinterpret_cast<CIDA *>(files);
	UINT i = 0;

	TRACE(TRACE_STATUS,"Drop_SHELLIDLIST");

	ZeroMemory(&ExecInfo,sizeof(ExecInfo));
	ExecInfo.cbSize = sizeof(ExecInfo);
	ExecInfo.nShow = SW_SHOWNORMAL;
	ExecInfo.hwnd = GlobalData.hMainWnd;

#ifndef USE_PIDL

	IShellFolder *psfDesktop;
	IShellFolder *psfParent;
	STRRET strretName;
	char Name[MAX_LINE_LENGTH];

	/* we want the name as a regular C string if possible*/ 
	strretName.uType = STRRET_CSTR;
	ExecInfo.lpFile = Name;

	/* get the IShellFolder interface for the parent folder */ 
	SHGetDesktopFolder(&psfDesktop);
	psfDesktop->BindToObject(HIDA_GetPIDLParent(pCIDA),NULL,IID_IShellFolder,(void**)&psfParent);
	psfDesktop->Release();

	if (psfParent==NULL)
	{
		TRACE(TRACE_WARNING,"Unable to get IShellFolder interface for parent.");
		return;
	}

	while (i<pCIDA->cidl)
	{
		if (psfParent->GetDisplayNameOf(HIDA_GetPIDLItem(pCIDA,i),SHGDN_NORMAL|SHGDN_FORPARSING,&strretName)
			==NOERROR)
		{
			/* Unfortunately, my shlwapi.h does not appear to have this function. */ 
			/* Since I don't know how else to get the string, I guess I'll have   */ 
			/* to go back to using PIDLs, since it almost works right.            */ 
			if (StrRetToBuf(&strretName,HIDA_GetPIDLItem(pCIDA,i),Name,MAX_LINE_LENGTH)
				==NOERROR)
			{
				ShellExecuteEx(&ExecInfo);
			}
		}
		++i;
	}

	psfParent->Release();

#else /* USE_PIDL */ 

	LPMALLOC pMalloc;

	ExecInfo.fMask = SEE_MASK_IDLIST;

	SHGetMalloc(&pMalloc);

/*		Multimon should use
					SEE_MASK_HMONITOR
					ExecInfo.hMonitor
*/

	while (i<pCIDA->cidl)
	{
		// stupid shell gives me a folder and a relative pidl
		// where i really need an absolute pidl
		ExecInfo.lpIDList = AppendPIDL(pMalloc,
			HIDA_GetPIDLParent(pCIDA),HIDA_GetPIDLItem(pCIDA,i));

		// if you send a pidl that corresponds to a folder to ShellExecuteEx
		// it will open that folder, then pop up a message saying that it can't
		// open '(null)'   For some reason, it doesn't ignore the lpFile 
		// variable in this case and I don't know how to make it.
		//SetExplorerIfNecessary(&ExecInfo);
		ShellExecuteEx(&ExecInfo);

		pMalloc->Free(ExecInfo.lpIDList);

		++i;
	}

	pMalloc->Release();
#endif
}

void VWMDropTarget::Drop_HDROP(void *files)
{
	DROPFILES *df = reinterpret_cast<DROPFILES *>(files);
	wchar_t *names;
	char name[MAX_PATH_LENGTH];

	TRACE(TRACE_STATUS,"Drop_HDROP");

	// damn unicode
	names = reinterpret_cast<wchar_t *>(reinterpret_cast<BYTE *>(files)+df->pFiles);
	while (*names!=0)
	{
		WideCharToMultiByte(CP_ACP,0,names,-1,name,MAX_PATH_LENGTH,NULL,NULL);

		TRACE(TRACE_STATUS,name);
		LSExecute(NULL, name, SW_SHOWNORMAL);

		while (*++names!=0) ;
		++names;
	}
}

void VWMDropTarget::Drop_INETURL(void *text)
{
	char *p = reinterpret_cast<char *>(text);
	char name[MAX_PATH_LENGTH]={0};
	TRACE(TRACE_STATUS,"Drop_INETURL");

	if (p[1]==0)
	{
		// assume urls are never 1 byte long
		// so this one's in unicode
		WideCharToMultiByte(CP_ACP,0,reinterpret_cast<wchar_t *>(text),-1,name,MAX_PATH_LENGTH,NULL,NULL);
		p = name;
	}
	else
	{
		p = reinterpret_cast<char *>(text);
	}

	LSExecute(NULL,p,SW_SHOWNORMAL);
}

void VWMDropTarget::Drop_SHELLURL(void *text)
{
	TRACE(TRACE_STATUS,"Drop_SHELLURL");
	Drop_INETURL(text);
}

void VWMDropTarget::Drop_TEXT(void *text)
{
	char *p = reinterpret_cast<char *>(text);
	char *eol;
	char nl = 13;

	TRACE(TRACE_STATUS,"Drop_TEXT");

	// parse it line-for-line and execute it
	
	eol = strchr(p,nl);
	if (!eol) { nl=10; eol = strchr(p,nl); }
	while (eol)
	{
		*eol = 0;

		LSExecute(GlobalData.hMainWnd,p,SW_SHOWNORMAL);

		p = eol+1;
		while (*p==13 || *p==10) ++p;

		eol = strchr(p,nl);
	}
	LSExecute(GlobalData.hMainWnd,p,SW_SHOWNORMAL);
}

void VWMDropTarget::Drop_UNICODETEXT(void *text)
{
	char buffer[MAX_LINE_LENGTH];

	TRACE(TRACE_STATUS,"Drop_UNICODETEXT");

	WideCharToMultiByte(CP_ACP,0,reinterpret_cast<wchar_t *>(text),-1,buffer,
		MAX_LINE_LENGTH,NULL,NULL);

	Drop_TEXT(buffer);
}

#endif
