/*
  RabidVWM LiteStep Module - version 1.0
	Common Painting Functions

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
		This file includes generic painting code, such as looping over
		all the windows, drawing icons, and combining back, windows, and
		select layers.  All alpha transparency and settings that are common
		to the different painting modes is alo handled here.

		Specific code for drawing with solid colors, bitmaps, and snapshots
		are in separate files.  (and switched in readconfig with pointers
		to avoid having to check settings during each repaint)

		To paint the miniwindows, first the window loop init function is
		called for the specific type, then the window and titlebar (if
		appropriate) painters are called for each window, from the bottom
		to the top, then finally the cleanup function is called.
*/

#include "RabidVWM.h"
#include "bitfiddle.h"

// layer painters
void PaintLayer(void (*LayerPainter)(PixelBlock &,HDC),HDC dc);
void PaintBack(PixelBlock &,HDC);
void PaintSel(PixelBlock &,HDC);
void PaintWnd(PixelBlock &,HDC);


// stolen from tasks, since it works better
// probably was missing windows that didn't have both large & small icons
HICON getIconFromWindow(HWND hwnd, BOOL bBigIcon)
{/*This code is a mix from re5ource's Popups and Fahim's lsTasks*/
	// now has a little ckVWM mixed in
	HICON a=NULL;
	if (!bBigIcon){
		SendMessageTimeout(hwnd, WM_GETICON, ICON_SMALL, 0, SMTO_ABORTIFHUNG, 1000, (unsigned long*)&a);
		if (!a) a = (HICON)GetClassLong(hwnd, GCL_HICONSM);
		if (!a) SendMessageTimeout(hwnd, WM_GETICON, ICON_BIG, 0, SMTO_ABORTIFHUNG, 1000, (unsigned long*)&a);
		if (!a) a = (HICON)GetClassLong(hwnd, GCL_HICON);
		if (!a) SendMessageTimeout(hwnd, WM_QUERYDRAGICON, 0, 0, SMTO_ABORTIFHUNG, 1000, (unsigned long*)&a);
	}else{
		SendMessageTimeout(hwnd, WM_GETICON, ICON_BIG, 0, SMTO_ABORTIFHUNG, 1000, (unsigned long*)&a);
		if (!a) a = (HICON)GetClassLong(hwnd, GCL_HICON);
		if (!a) SendMessageTimeout(hwnd, WM_QUERYDRAGICON, 0, 0, SMTO_ABORTIFHUNG, 1000, (unsigned long*)&a);
		if (!a) SendMessageTimeout(hwnd, WM_GETICON, ICON_SMALL, 0, SMTO_ABORTIFHUNG, 1000, (unsigned long*)&a);
		if (!a) a = (HICON)GetClassLong(hwnd, GCL_HICONSM);
	}
	/* no default icon
	if (!a)
		a = LoadIcon(NULL, MAKEINTRESOURCE(IDI_APPLICATION));
		*/
	return a;
}

// ***************************************** Repaint

void Repaint(void)
{
	HDC pdc,mdc;
	HBITMAP obm;

	if (!IsWindowVisible(GlobalData.hMainWnd) ||
		/* do we need to check the parent? */ 
		(GlobalData.hParentWnd!=NULL && !GlobalData.bAlwaysOnTop
			&& !IsWindowVisible(GlobalData.hParentWnd)))
	{
		/* never draw if the vwm is hidden */ 
#if !defined(NDEBUG) && defined(TRACE_FILE)
		if (!GlobalData.bConstantUpdate)
			TRACE(TRACE_STATUS,"Repaint: VWM is hidden.");
#endif
		return;
	}

#ifdef VWM_STATS
	int start = GetTickCount();
#endif

	if (GlobalData.bTransparent)
	{
		PixelBlock pb(GlobalData.VWMSize.cx,GlobalData.VWMSize.cy);
		pdc = GetDC(NULL);
		mdc = CreateCompatibleDC(pdc);
		ReleaseDC(NULL,pdc);
		obm = (HBITMAP)SelectObject(mdc,pb.Handle());

		/* fill it with TRANSPARENT_COLOR first */ 
		{
			PixelBlockIterator pbi(pb);
			while (!pbi.Done()) { pbi.Set(TRANSPARENT_COLOR); ++pbi; }
		}

		PaintBack(pb,mdc);
		PaintSel(pb,mdc);
		PaintWnd(pb,mdc);

		/* premultiply the alpha values */ 
		{
			PixelBlockIterator pbi(pb);
			while (!pbi.Done()) { pbi.MultiplyAlpha(); ++pbi; }
		}

		POINT ptSRC = {0,0};
		BLENDFUNCTION bl;
		bl.BlendOp = AC_SRC_OVER;
		bl.BlendFlags = 0;
		bl.AlphaFormat = AC_SRC_ALPHA;
		bl.SourceConstantAlpha = 255;
		dUpdateLayeredWindow(GlobalData.hMainWnd,
			NULL,&GlobalData.VWMPos,&GlobalData.VWMSize,
			mdc,&ptSRC,
			TRANSPARENT_COLOR,&bl,ULW_ALPHA/*|ULW_COLORKEY*/);

		SelectObject(mdc,obm);
		DeleteDC(mdc);
	}
  else /* VWMTransparent */ 
  {
	  pdc = GetDC(GlobalData.hMainWnd);
	  mdc = CreateCompatibleDC(pdc);
	  StatGDIAdd(DC);
	  obm = (HBITMAP)SelectObject(mdc,GlobalData.bmBuffer);
	  ReleaseDC(GlobalData.hMainWnd,pdc);

	  /* fill it with TRANSPARENT_COLOR first */ 
	  if (GlobalData.bUseTransBlt || GlobalData.cBack.Alpha!=255) {
		  if (GlobalData.bUseFakeTrans && GlobalData.bmTransparent!=NULL) {
			  /* somehow repainting seems slower now, but it shouldn't be */ 
			  pdc = CreateCompatibleDC(mdc);
			  GlobalData.bmTransparent = (HBITMAP)SelectObject(pdc,GlobalData.bmTransparent);

			  BitBlt(mdc,0,0,GlobalData.VWMSize.cx,GlobalData.VWMSize.cy,pdc,0,0,SRCCOPY);

			  GlobalData.bmTransparent = (HBITMAP)SelectObject(pdc,GlobalData.bmTransparent);
			  DeleteDC(pdc);
		  }
		  else
		  {
			  HBRUSH brTrans,brOld;
			  brTrans = CreateSolidBrush(TRANSPARENT_COLOR);
			  StatGDIAdd(Brush);
			  brOld = (HBRUSH)SelectObject(mdc,brTrans);

			  Rectangle(mdc,-1,-1,GlobalData.VWMSize.cx+1,GlobalData.VWMSize.cy+1);

			  SelectObject(mdc,brOld);
			  DeleteObject(brTrans);
			  StatGDIDel(Brush);
		  }
	  }

	  ASSERT(GlobalData.PaintVWM!=NULL);

	  PaintLayer(PaintBack,mdc);
	  PaintLayer(PaintSel,mdc);
	  PaintLayer(PaintWnd,mdc);

	  SelectObject(mdc,obm);
	  DeleteDC(mdc);
	  StatGDIDel(DC);

	  GlobalData.bChangeState = false;
	  GlobalData.bDelaySnapshot = false;

	  if (!GlobalData.bTransparent)
		  InvalidateRect(GlobalData.hMainWnd, NULL, FALSE);

  } /* VWMTransparent */ 

#ifdef VWM_STATS
	VWMStats.PaintTime += GetTickCount()-start;
#endif

  if (GlobalData.nSetupPass==0)
    AtVarExec(GlobalData.OnRepaint,true,true);
}

// common paint routines

// slice the area into desk areas and paint them
void PaintVWM_Generic(HDC dc,void (*PaintDesk)(HDC,s_DeskData *,RECT *))
{
	int x,y,j;
	RECT r;
	int left,top,width,height;

	left = GlobalData.rDeskBorder.left;
	top = GlobalData.rDeskBorder.top;
	width = GlobalData.VWMSize.cx - left - GlobalData.rDeskBorder.right;
	height = GlobalData.VWMSize.cy - top - GlobalData.rDeskBorder.bottom;

	width = GlobalData.nDesksX * GlobalData.VWMDeskSize.cx;
	height = GlobalData.nDesksY * GlobalData.VWMDeskSize.cy;

	r.top = top+height; /* these will get shifted to bottom right before use */ 
	for (y=GlobalData.nDesksY,j=GlobalData.nDesks;y>0;y--,j-=GlobalData.nDesksX) {
		/* calculate new vertical coords */ 
		r.bottom = r.top;
		r.top = r.top-GlobalData.VWMDeskSize.cy;

		r.left = left+width;
		for (x=GlobalData.nDesksX;x>0;x--) {
			/* calculate new horiz coords */ 
			r.right = r.left;
			r.left = r.left-GlobalData.VWMDeskSize.cx;

			/* draw the desk              */ 
			if (!GlobalData.bNoDrawSelBack ||
					GlobalData.CurrentDesk != (x+j-1-GlobalData.nDesksX))
				PaintDesk(dc,&GlobalData.Desks[x+j-1-GlobalData.nDesksX],&r);
		}
	}
}

// paint one of the three layers onto the mdc
void PaintLayer(void (*LayerPainter)(PixelBlock &,HDC),HDC mdc)
{
  ASSERT(EXECABLE(LayerPainter));

	HBITMAP bmOld;
	HDC dc = CreateCompatibleDC(mdc);
	PixelBlock pb(GlobalData.VWMSize.cx,GlobalData.VWMSize.cy);
	StatGDIAdd(DC);
	bmOld = (HBITMAP)SelectObject(dc,pb.Handle());

	{
		/* fill it with TRANSPARENT_COLOR first */ 
		PixelBlockIterator pbi(pb);
		while (!pbi.Done()) { pbi.Set(TRANSPARENT_COLOR); ++pbi; }
	}

	LayerPainter(pb,dc);

	if (dAlphaBlend!=NULL) {
		/* premultiply the alpha values */ 
		{
			PixelBlockIterator pbi(pb);

			while (!pbi.Done())
			{
				pbi.MultiplyAlpha();
				++pbi;
			}
		}
		{
			int width = GlobalData.VWMSize.cx;
			int height= GlobalData.VWMSize.cy;
			BLENDFUNCTION bl;
			bl.BlendOp = AC_SRC_OVER;
			bl.BlendFlags = 0;
			bl.AlphaFormat = AC_SRC_ALPHA;
			bl.SourceConstantAlpha = 255;
			dAlphaBlend(mdc,0,0,width,height,dc,0,0,width,height,bl);
		}
	} else {
		CTRACE(TRACE_WARNING,dAlphaBlend==NULL,"AlphaBlend not linked.");
		TransparentBltLS(mdc,0,0,GlobalData.VWMSize.cx,GlobalData.VWMSize.cy,dc,0,0,TRANSPARENT_COLOR);
	}

	SelectObject(dc,bmOld);
	DeleteDC(dc);
	StatGDIDel(DC);
}

// paint background layer
void PaintBack(PixelBlock &pb,HDC dc)
{
	if (GlobalData.PaintVWM==NULL) return;
	if (GlobalData.cBack.Alpha==0) return;

	GlobalData.PaintVWM(dc);

	if (dAlphaBlend!=NULL) {
		/* set the alpha for non-transparent */ 
		/* bits, but don't premultiply it    */ 
		{
			int alpha;
			PixelBlockIterator pbi(pb);

			alpha = GlobalData.cBack.Alpha;

			while (!pbi.Done()) {
				if (pbi.Get()!=TRANSPARENT_COLOR && pbi.Alpha()==0)
					pbi.SetAlpha(alpha);
				++pbi;
			}
		}
	}
}

// paint select layer
void PaintSel(PixelBlock &pb,HDC dc)
{
	RECT r;

	if (GlobalData.PaintVWM==NULL) return;
	if (GlobalData.cSelect.Alpha==0) return;

	/* select the current desk                 */ 
	GetDeskVWMRect(GlobalData.CurrentDesk,&r);
	/* now adjust for select expand border     */ 
	r.left -= GlobalData.rSelExpand.left;
	r.top -= GlobalData.rSelExpand.top;
	r.right += GlobalData.rSelExpand.right;
	r.bottom += GlobalData.rSelExpand.bottom;

	/* call the specific selected desk painter */ 
	GlobalData.PaintSelect(dc,&GlobalData.Desks[GlobalData.CurrentDesk],&r);

	if (dAlphaBlend!=NULL) {
		{
			int alpha;
			PixelBlockIterator pbi(pb);

			alpha = GlobalData.cSelect.Alpha;

			while (!pbi.Done()) {
				if (pbi.Get()!=TRANSPARENT_COLOR && pbi.Alpha()==0)
					pbi.SetAlpha(alpha);
				++pbi;
			}
		}
	}
}

// paint the window layer
void PaintWnd(PixelBlock &pb,HDC dc)
{
	char store[PAINTWND_BUFFER_SIZE/sizeof(char)];
	int i;
#if !defined(NDEBUG)
	int loopdetect = GlobalData.nKnownWindows;
#endif
	RECT rFull,rPart;
	s_WinData *pData;

	/* nothing to paint */ 
	if (GlobalData.bNoWindows || GlobalData.nKnownWindows==0) return;

	ASSERT(READABLE(GlobalData.Painter,sizeof(s_PaintWnd)));
	ASSERT(EXECABLE(GlobalData.Painter->BeginPaintWnd));
	ASSERT(EXECABLE(GlobalData.Painter->PaintWnd));
	/* if I can't draw a titlebar, it'll be null, but then better */ 
	/* clear the flag when you set the painter                    */ 
	ASSERT(!GlobalData.bTitlebars || EXECABLE(GlobalData.Painter->PaintWndTitle));
	ASSERT(EXECABLE(GlobalData.Painter->EndPaintWnd));

	INITTRACE(TRACE_STATUS,"BeginPaintWnd");
	GlobalData.Painter->BeginPaintWnd(dc,pb,&store);

	/* paint windows in order */ 
	i = GlobalData.FirstWindow;
	while (i>=0) {

		ASSERT(i>=0);
		ASSERT(i<GlobalData.nKnownWindows);
		pData = &GlobalData.KnownWindows[i];

		if (ShouldDrawWnd(pData))
		{
			rFull = pData->Mini;

			rPart = rFull;

			if (!GlobalData.bTitlebars ||
				(rFull.bottom-rFull.top)<=GlobalData.TitlebarHeight ||
				(GlobalData.bTitleForTitle && !HasTitleBar(pData->hWnd))
				) {

				GlobalData.Painter->PaintWnd(dc,pb,&store,pData,&rFull,&rPart);

				rPart.top += GlobalData.IconBorder.top;
			} else {

				rPart.top += GlobalData.TitlebarHeight - 2;
				GlobalData.Painter->PaintWnd(dc,pb,&store,pData,&rFull,&rPart);

				rPart.bottom = rPart.top + 2;
				rPart.top = rFull.top;
				GlobalData.Painter->PaintWndTitle(dc,pb,&store,pData,&rFull,&rPart);

				/* set up rPart for drawing icons */ 
				rPart.top = rPart.bottom;
				rPart.bottom = rFull.bottom;

				if (GlobalData.bForceIconBorderTop)
					rPart.top = rFull.top + GlobalData.IconBorder.top;
			}
			/* rPart.top depends on titlebar */ 
			rPart.left+=GlobalData.IconBorder.left;
			rPart.right-=GlobalData.IconBorder.right;
			rPart.bottom-=GlobalData.IconBorder.bottom;

			if (GlobalData.bShowIcons) {
				/* draw the icon centerred in rPart */ 
				int iconsz = GlobalData.IconSize;
				int iconw = rPart.right-rPart.left;
				int iconh = rPart.bottom-rPart.top;

				if (GlobalData.bScaleIconToFit) {
					if (iconw<iconsz) iconsz = iconw;
					if (iconh<iconsz) iconsz = iconh;
				}

				if (!GlobalData.bScaleIconToFit || iconsz>=GlobalData.MinIconSize)
				{
					/* get the icon if the cached one is no good */ 
					if (pData->LastIconSize!=iconsz || pData->hIcon==NULL) {
						/* small icon if scaling to <= 16 */ 
						if (iconsz <= 16)
							pData->hIcon = getIconFromWindow(pData->hWnd, FALSE);
						/* else large icon                */ 
						else
							pData->hIcon = getIconFromWindow(pData->hWnd, TRUE);
						pData->LastIconSize = iconsz;
					}
					/* draw it */ 
					if (
						!DrawIconEx(dc,
							rPart.left+(iconw-iconsz)/2,
							rPart.top+(iconh-iconsz)/2,
							pData->hIcon,
							iconsz, iconsz, 0, NULL, DI_NORMAL)
						)
					/* clear the icon if it failed so we'll try again to get it */ 
						pData->hIcon = NULL;
				}
			}
			{
				int alpha;
				PixelBlockIterator pbi(pb,rFull);

				if ((pData->bFlash && GlobalData.bFlashState) || pData->bFlashOn)
					alpha = GlobalData.cFlash.Alpha;
				else alpha = pData->Alpha;

				while (!pbi.Done()) {
					if (pbi.Get()!=TRANSPARENT_COLOR && pbi.Alpha()==0)
						pbi.SetAlpha(alpha);
					++pbi;
				}
			}

			pData->bWasDrawn = true;
		}

		pData->OldArea = pData->Area;
		pData->bTitleChanged = false;
		pData->bPosChanged = false;
		pData->bSizeChanged = false;

		i = pData->PaintNext;
#if !defined(NDEBUG)
		/* see if we've drawn more windows than exist */ 
		if (--loopdetect<0) {
			TRACE(TRACE_ERROR,"Found loop in paint order");
			DumpWindowList();
			break;
		}
#endif
	}

	INITTRACE(TRACE_STATUS,"EndPaintWnd");
	GlobalData.Painter->EndPaintWnd(dc,pb,&store);

	if (GlobalData.bItsEaster) ShowEaster(&pb);
}
