/*
  RabidVWM LiteStep Module - version 1.0
	Implementation for Special * Window stuff

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/*
		Since DoWindowMatches is called during each window list update,
		these will be called for all the windows during each update,
		so they should be quick.
*/

#include "RabidVWM.h"

static HWND GetH_Parent(s_WinData *pData)
{
	return GetParent(pData->hWnd);
}

void Set_IgnoreTitle(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	TRACE(TRACE_STATUS,"Set_IgnoreTitle");
	pData->bIgnoreTitleChange = true;
}

void Clear_IgnoreTitle(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	pData->bIgnoreTitleChange = false;
}

void Set_Sticky(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	TRACE(TRACE_STATUS,"Set_Sticky");
	pData->bSticky = true;
}

void Clear_Sticky(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	pData->bSticky = false;
}

void Set_StartDesk(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	int nDesk,i;
	HWND h;

	TRACE(TRACE_STATUS,"Set_StartDesk");

	/* move to desk only if new window              */ 
	if (!pData->bNewWindow) return;

	/* can't do anything if no desk specified!      */ 
	if (pMatch->Args==NULL) return;

	/* get desk                                     */ 
	nDesk = GetBangDeskNo(pMatch->Args,NULL);

	/* if desk number specified was 0, try to guess */ 
  if (nDesk<0) { /* < because 1-based -> 0-based  */ 
		/* first try window's owner */ 
		h = GetWindow(pData->hWnd, GW_OWNER);
		if (h) {
			i = FindKnownWindow(h);
			if (GlobalData.KnownWindows[i].hWnd == h) {
				nDesk = GlobalData.KnownWindows[i].HomeDesk;
			}
		}
		if (nDesk<0) {
			/* then try window's parent */ 
			h = GetParent(pData->hWnd);
			if (h) {
				i = FindKnownWindow(h);
				if (GlobalData.KnownWindows[i].hWnd == h) {
					nDesk = GlobalData.KnownWindows[i].HomeDesk;
				}
			}
			if (nDesk<0) {
				/* then give up */ 
				return;
			}
		}
	}

	MoveWindowToDesk(pData,nDesk,NULL);
}

void Clear_StartDesk(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	/* nothing to do here */ 
	return;
}

void Set_Fix(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	TRACE(TRACE_STATUS,"Set_Fix");
	pData->bFix = true;
	return;
}

void Clear_Fix(s_WinData *pData, StarWindowMatchItem *pMatch)
{
	if (pData->bHidden && pData->bFix) {
		ShowWindow(pData->hWnd,SW_SHOWNA);
		pData->bHidden = false;
	}
	pData->bFix = false;
	return;
}

StarWindowMatchList StarLists[] = {
  /* Key,                    Item list, Set proc,         Clear proc,         handle adjust proc */ 
	{ "*VWMIgnoreTitleChange",			NULL, &Set_IgnoreTitle,	&Clear_IgnoreTitle,	NULL					},
	{ "*VWMSticky",									NULL, &Set_Sticky,			&Clear_Sticky,			NULL					},
	{ "*VWMStartWindowsOnDesktop",	NULL, &Set_StartDesk,		&Clear_StartDesk,		NULL					},
	{ "*VWMFix",										NULL, &Set_Fix,					&Clear_Fix,					&GetH_Parent	},
	{ NULL,													NULL, NULL,							NULL,								NULL					}
};

