/*
  RabidVWM LiteStep Module - version 1.0
	Generic * Window Code

  Copyright (C) 2002 Joshua Seagoe
 
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


#include "RabidVWM.h"
#include <stdlib.h>

#ifndef LSTOOLS_REV
// read one set of *list items, without lstools
// (should merge lstools and non-lstools versions, as in SkinBox)
static void ReadStarList(const char *key, StarWindowMatchList *list)
{
	FILE *f;
	StarWindowMatchItem *pItem;
	char *wndspec,*args;
	int i;
	DWORD flags;

	/* must be called from ReadConfig only! */ 
	ASSERTR(READABLE(g_pParseData,sizeof(s_ParseData)),;);
	ASSERTR(WRITEABLE(g_pParseData,sizeof(s_ParseData)),;);
	ASSERTR(WRITEABLE(list,sizeof(*list)),;);
	ASSERTR(READABLE(key,sizeof(char)),;);

#if !defined(NDEBUG) && defined(TRACE_FILE)
	int count = 0;
	{
		char buffer[256];
		sprintf(buffer,"Reading %s from %s",key,GlobalData.ConfigFile?GlobalData.ConfigFile:"NULL");
		TRACE(TRACE_STATUS,buffer);
	}
#endif

	list->pItem = NULL;

	f = LCOpen(GlobalData.ConfigFile);
	if (f)
	{
		while (LCReadNextConfig (f, key, g_pParseData->line, MAX_LINE_LENGTH))
		{
			TRACE(TRACE_STATUS,g_pParseData->line);

			*g_pParseData->one =
			*g_pParseData->two =
			*g_pParseData->three = 0;
			LCTokenize(g_pParseData->line, g_pParseData->tokens, 2, g_pParseData->three);

			wndspec = g_pParseData->two;
			args = g_pParseData->three;

			if (*wndspec=='#') {
				/* class or title flags  */ 
				flags = ParseWindowSpecFlags(wndspec);
				/* wndspec is next token */ 
				*g_pParseData->one =
				*g_pParseData->two = 0;
				LCTokenize(g_pParseData->three, g_pParseData->tokens, 1, g_pParseData->two);
				wndspec = g_pParseData->one;
				args = g_pParseData->two;

			} else {
				flags = MATCH_WND_TITLE | MATCH_WND_CLASS;
		    /* I would like to issue a warning of some sort here.         */ 
		    /* Not specifying the flags makes me look at *BOTH* class and */ 
		    /* title when the user should know which they're specifying.  */ 
        /* but it's perfectly legal, so that'd be too intrusive, imo. */ 
			}

			i = lstrlen(wndspec);
			if (i==0) continue;

			/* create item   */ 
			pItem = new StarWindowMatchItem;
			/* set match key */ 
			pItem->Key = new char[i+1];

			lstrcpy(pItem->Key,wndspec);
			pItem->MatchProcFlag = flags;
			/* set args      */ 
			i = lstrlen(args);
			if (i) {
				pItem->Args = new char[i+1];
				lstrcpy(pItem->Args,args);
			} else {
				pItem->Args = NULL;
			}

			pItem->pNext = list->pItem;

			/* update list   */ 
			list->pItem = pItem;
#if !defined(NDEBUG) && defined(TRACE_FILE)
			count++;
#endif
		}
		LCClose(f);
	}
#if !defined(NDEBUG) && defined(TRACE_FILE)
	{
		char buffer[256];
		sprintf(buffer,"%d items read.",count);
		TRACE(TRACE_STATUS,buffer);
	}
#endif
}

// free one set of *list items, without lstools
static void UnreadStarList(StarWindowMatchList *list)
{
	StarWindowMatchItem *pItem;

	ASSERTR(READABLE(list,sizeof(StarWindowMatchList)),;);
	ASSERTR(WRITEABLE(list,sizeof(StarWindowMatchList)),;);

#if !defined(NDEBUG) && defined(TRACE_FILE)
	int count = 0;
	{
		char buffer[256];
		sprintf(buffer,"Releasing %s.",list->Key);
		TRACE(TRACE_STATUS,buffer);
	}
#endif

	while ((pItem = list->pItem)) {
		list->pItem = pItem->pNext;
		ASSERT(pItem!=NULL);
		ASSERT(pItem->Key!=NULL);
		delete[] pItem->Key;
		if (pItem->Args) delete[] pItem->Args;
		delete pItem;
#if !defined(NDEBUG) && defined(TRACE_FILE)
		count++;
#endif
	}

#if !defined(NDEBUG) && defined(TRACE_FILE)
	{
		char buffer[256];
		sprintf(buffer,"%d items released.",count);
		TRACE(TRACE_STATUS,buffer);
	}
#endif
#ifndef NDEBUG
	list->pItem = NULL;
#endif
}
#else /* LSTOOLS_REV */ 

// read one *list item, with LSTOOLS
static int ReadStarListProc(void *pElem,const char *line,DWORD param)
{
	StarWindowMatchItem *pItem = (StarWindowMatchItem *)pElem;
	char *wndspec,*args;
	int i;
	DWORD flags;

	*g_pParseData->one =
	*g_pParseData->two =
	*g_pParseData->three = 0;
	lstrcpy(g_pParseData->line,line);
	LCTokenize(g_pParseData->line, g_pParseData->tokens, 1, g_pParseData->three);

	wndspec = g_pParseData->one;
	args = g_pParseData->three;

	if (*wndspec=='#') {
		/* class or title flags  */ 
		flags = ParseWindowSpecFlags(wndspec);
		/* wndspec is next token */ 
		*g_pParseData->one =
		*g_pParseData->two = 0;
		LCTokenize(g_pParseData->three, g_pParseData->tokens, 1, g_pParseData->two);
		wndspec = g_pParseData->one;
		args = g_pParseData->two;

	} else {
		flags = MATCH_WND_TITLE | MATCH_WND_CLASS;
		/* I would like to issue a warning of some sort here.         */ 
		/* Not specifying the flags makes me look at *BOTH* class and */ 
		/* title when the user should know which they're specifying.  */ 
    /* but it's perfectly legal, so that'd be too intrusive, imo. */ 
	}

	i = lstrlen(wndspec);
	if (i==0) return PSL_FAIL|PSL_CONTINUE;

	/* set match key */ 
	pItem->Key = new char[i+1];

	lstrcpy(pItem->Key,wndspec);
	pItem->MatchProcFlag = flags;
	/* set args      */ 
	i = lstrlen(args);
	if (i) {
		pItem->Args = new char[i+1];
		lstrcpy(pItem->Args,args);
	} else {
		pItem->Args = NULL;
	}

	return PSL_OK|PSL_CONTINUE;
}

// read one set of *items, with LSTOOLS
static void ReadStarList(const char *key, StarWindowMatchList *list)
{
	FILE *f = LCOpen(GlobalData.ConfigFile);
  ASSERT(READABLE(key,sizeof(char)));
  ASSERT(READABLE(list,sizeof(StarWindowMatchList))));
  ASSERT(WRITEABLE(list,sizeof(StarWindowMatchList))));
	LSParseStarList(f,key,&ReadStarListProc,&list->pItem,sizeof(StarWindowMatchItem)/4,NULL);
	LCClose(f);
}

// free one *item, with LSTOOLS
static void FreeStarListProc(void *pElem,DWORD param)
{
	StarWindowMatchItem *pItem = (StarWindowMatchItem *)pElem;
  ASSERT(READABLE(pItem,sizeof(StarWindowMatchItem)));
  ASSERT(WRITEABLE(pItem,sizeof(StarWindowMatchItem)));

	delete[] pItem->Key;
	if (pItem->Args) delete[] pItem->Args;
}

// free one set of *items, with LSTOOLS
static void UnreadStarList(StarWindowMatchList *list)
{
  ASSERT(READABLE(list,sizeof(StarWindowMatchList)));
	LSFreeStarList(&FreeStarListProc,&list->pItem,NULL);
}
#endif /* LSTOOLS_REV */ 

// compare a window to every item in a starlist and apply its Set and Clear functions
static bool DoWindowMatchList(
	s_WinData *pData,
	StarWindowMatchList *list,
	const char *WndTitle,
	const char *WndClass
)
{
	StarWindowMatchItem *pItem;
	char bT[256],bC[256];
	HWND h;

	ASSERTR(READABLE(pData,sizeof(s_WinData)),false);
	ASSERTR(WRITEABLE(pData,sizeof(s_WinData)),false);
	ASSERTR(READABLE(list,sizeof(StarWindowMatchList)),false);
	ASSERTR(READABLE(WndTitle,sizeof(char)),false);
	ASSERTR(READABLE(WndClass,sizeof(char)),false);

	if (list->GetH != NULL && list->pItem!=NULL) {
		/* use the title/class from a different window       */ 
		h = list->GetH(pData);
		GetWindowText(h, bT, 256); /* this buffer is shorter */ 
		GetClassName(h, bC, 256);  /* this is the same, tho  */ 
		/* now, USE them                                     */ 
		WndTitle = bT;
		WndClass = bC;
	}

  /* Shouldn't have to check title and class every time... */ 
  /* We detect title changes, so it's not called for every */ 
  /* update, but the window class will never change, so we */ 
  /* only need to check it once for each window.           */ 
  /* Oh well, someday this might be something to fix.      */ 

	pItem = list->pItem;
	while (pItem) {
		ASSERTR(pItem->Key!=NULL,false);
		if ((pItem->MatchProcFlag&MATCH_WND_TITLE && match(pItem->Key,WndTitle)))
		{
      /* Set should check if it's already been called for */ 
      /* this window if it's a time-consuming operation,  */ 
      /* or if it should only happen once for the window. */ 
			list->Set(pData,pItem);
			return true;
		}
		if ((pItem->MatchProcFlag&MATCH_WND_CLASS && match(pItem->Key,WndClass)))
		{
			list->Set(pData,pItem);
			return true;
		}
		pItem = pItem->pNext;
	}
  /* clear it if it's not being set */ 
	list->Clear(pData,NULL);
	return false;
}

extern StarWindowMatchList StarLists[];

// this checks every known StarWindowMatchList
void DoWindowMatches(s_WinData *pData, const char *title)
{
	char WndClass[256];
	StarWindowMatchList *pList;
#ifdef VWM_STATS
	int start = GetTickCount();
#endif

	ASSERTR(READABLE(pData,sizeof(s_WinData)),;);
	ASSERTR(WRITEABLE(pData,sizeof(s_WinData)),;);
	ASSERTR(READABLE(title,sizeof(char)),;);

	GetClassName(pData->hWnd, WndClass, 256);

	pList = StarLists;
	while (pList->Key != NULL) {
		DoWindowMatchList(pData, pList, title, WndClass);
		pList++;
	}

#ifdef VWM_STATS
	VWMStats.StarListTime += GetTickCount()-start;
#endif
}


// read and clean up all star lists

void ReadStarLists(void)
{
	StarWindowMatchList *pList;

	pList = StarLists;
	while (pList->Key != NULL) {
		ReadStarList(pList->Key, pList);
		pList++;
	}
}

void UnreadStarLists(void)
{
	StarWindowMatchList *pList;

	pList = StarLists;
	while (pList->Key != NULL) {
		UnreadStarList(pList);
		pList++;
	}
}

